from __future__ import annotations

from abc import abstractmethod
from collections.abc import Callable, Iterator, Sequence
from typing import (
    Any,
    Generic,
    NewType,
    NoReturn,
    Protocol,
    TypeVar,
    cast,
    get_args,
    get_type_hints,
    runtime_checkable,
)

from ._device import Device, DeviceConnector, DeviceVector
from ._signal import Signal, SignalX
from ._signal_backend import SignalBackend, SignalDatatype
from ._utils import get_origin_class

SignalBackendT = TypeVar("SignalBackendT", bound=SignalBackend)
DeviceConnectorT = TypeVar("DeviceConnectorT", bound=DeviceConnector)
# Unique name possibly with trailing understore, the attribute name on the Device
UniqueName = NewType("UniqueName", str)
# Logical name without trailing underscore, the name in the control system
LogicalName = NewType("LogicalName", str)


def _get_datatype(annotation: Any) -> type | None:
    """Return int from SignalRW[int]."""
    args = get_args(annotation)
    if len(args) == 1 and get_origin_class(args[0]):
        return args[0]


def _logical(name: UniqueName) -> LogicalName:
    return LogicalName(name.rstrip("_"))


@runtime_checkable
class DeviceAnnotation(Protocol):
    @abstractmethod
    def __call__(self, parent: Device, child: Device): ...


class DeviceFiller(Generic[SignalBackendT, DeviceConnectorT]):
    def __init__(
        self,
        device: Device,
        signal_backend_factory: Callable[[type[SignalDatatype] | None], SignalBackendT],
        device_connector_factory: Callable[[], DeviceConnectorT],
    ):
        self._device = device
        self._signal_backend_factory = signal_backend_factory
        self._device_connector_factory = device_connector_factory
        # Annotations stored ready for the creation phase
        self._uncreated_signals: dict[UniqueName, type[Signal]] = {}
        self._uncreated_devices: dict[UniqueName, type[Device]] = {}
        self._extras: dict[UniqueName, Sequence[Any]] = {}
        self._signal_datatype: dict[LogicalName, type | None] = {}
        self._vector_device_type: dict[LogicalName, type[Device] | None] = {}
        # Backends and Connectors stored ready for the connection phase
        self._unfilled_backends: dict[
            LogicalName, tuple[SignalBackendT, type[Signal]]
        ] = {}
        self._unfilled_connectors: dict[LogicalName, DeviceConnectorT] = {}
        # Once they are filled they go here in case we reconnect
        self._filled_backends: dict[
            LogicalName, tuple[SignalBackendT, type[Signal]]
        ] = {}
        self._filled_connectors: dict[LogicalName, DeviceConnectorT] = {}
        self._scan_for_annotations()

    def _raise(self, name: str, error: str) -> NoReturn:
        raise TypeError(f"{type(self._device).__name__}.{name}: {error}")

    def _store_signal_datatype(self, name: UniqueName, annotation: Any):
        origin = get_origin_class(annotation)
        datatype = _get_datatype(annotation)
        if origin == SignalX:
            # SignalX doesn't need datatype
            self._signal_datatype[_logical(name)] = None
        elif origin and issubclass(origin, Signal) and datatype:
            # All other Signals need one
            self._signal_datatype[_logical(name)] = datatype
        else:
            # Not recognized
            self._raise(
                name,
                f"Expected SignalX or SignalR/W/RW[type], got {annotation}",
            )

    def _scan_for_annotations(self):
        # Get type hints on the class, not the instance
        # https://github.com/python/cpython/issues/124840
        cls = type(self._device)
        # Get hints without Annotated for determining types
        hints = get_type_hints(cls)
        # Get hints with Annotated for wrapping signals and backends
        extra_hints = get_type_hints(cls, include_extras=True)
        for attr_name, annotation in hints.items():
            name = UniqueName(attr_name)
            origin = get_origin_class(annotation)
            if (
                name == "parent"
                or name.startswith("_")
                or not origin
                or not issubclass(origin, Device)
            ):
                # Ignore any child that is not a public Device
                continue
            self._extras[name] = getattr(extra_hints[attr_name], "__metadata__", ())
            if issubclass(origin, Signal):
                self._store_signal_datatype(name, annotation)
                self._uncreated_signals[name] = origin
            elif origin == DeviceVector:
                child_type = _get_datatype(annotation)
                child_origin = get_origin_class(child_type)
                if child_origin is None or not issubclass(child_origin, Device):
                    self._raise(
                        name,
                        f"Expected DeviceVector[SomeDevice], got {annotation}",
                    )
                if issubclass(child_origin, Signal):
                    self._store_signal_datatype(name, child_type)
                self._vector_device_type[_logical(name)] = child_origin
                setattr(self._device, name, DeviceVector({}))
            else:
                self._uncreated_devices[name] = origin

    def check_created(self):
        uncreated = sorted(set(self._uncreated_signals).union(self._uncreated_devices))
        if uncreated:
            raise RuntimeError(
                f"{self._device.name}: {uncreated} have not been created yet"
            )

    def create_signals_from_annotations(
        self,
        filled=True,
    ) -> Iterator[tuple[SignalBackendT, list[Any]]]:
        for name in list(self._uncreated_signals):
            child_type = self._uncreated_signals.pop(name)
            backend = self._signal_backend_factory(
                self._signal_datatype[_logical(name)]
            )
            extras = list(self._extras[name])
            yield backend, extras
            signal = child_type(backend)
            for anno in extras:
                assert isinstance(anno, DeviceAnnotation), anno
                anno(self._device, signal)
            setattr(self._device, name, signal)
            dest = self._filled_backends if filled else self._unfilled_backends
            dest[_logical(name)] = (backend, child_type)

    def create_devices_from_annotations(
        self,
        filled=True,
    ) -> Iterator[tuple[DeviceConnectorT, list[Any]]]:
        for name in list(self._uncreated_devices):
            child_type = self._uncreated_devices.pop(name)
            connector = self._device_connector_factory()
            extras = list(self._extras[name])
            yield connector, extras
            device = child_type(connector=connector)
            for anno in extras:
                assert isinstance(anno, DeviceAnnotation), anno
                anno(self._device, device)
            setattr(self._device, name, device)
            dest = self._filled_connectors if filled else self._unfilled_connectors
            dest[_logical(name)] = connector

    def create_device_vector_entries_to_mock(self, num: int):
        for name, cls in self._vector_device_type.items():
            assert cls, "Shouldn't happen"
            for i in range(1, num + 1):
                if issubclass(cls, Signal):
                    self.fill_child_signal(name, cls, i)
                elif issubclass(cls, Device):
                    self.fill_child_device(name, cls, i)
                else:
                    self._raise(name, f"Can't make {cls}")

    def check_filled(self, source: str):
        unfilled = sorted(set(self._unfilled_connectors).union(self._unfilled_backends))
        if unfilled:
            raise RuntimeError(
                f"{self._device.name}: cannot provision {unfilled} from {source}"
            )

    def _ensure_device_vector(self, name: LogicalName) -> DeviceVector:
        if not hasattr(self._device, name):
            # We have no type hints, so use whatever we are told
            self._vector_device_type[name] = None
            setattr(self._device, name, DeviceVector({}))
        vector = getattr(self._device, name)
        if not isinstance(vector, DeviceVector):
            self._raise(name, f"Expected DeviceVector, got {vector}")
        return vector

    def fill_child_signal(
        self,
        name: str,
        signal_type: type[Signal],
        vector_index: int | None = None,
    ) -> SignalBackendT:
        name = cast(LogicalName, name)
        if name in self._unfilled_backends:
            # We made it above
            backend, expected_signal_type = self._unfilled_backends.pop(name)
            self._filled_backends[name] = backend, expected_signal_type
        elif name in self._filled_backends:
            # We made it and filled it so return for validation
            backend, expected_signal_type = self._filled_backends[name]
        elif vector_index:
            # We need to add a new entry to a DeviceVector
            vector = self._ensure_device_vector(name)
            backend = self._signal_backend_factory(self._signal_datatype.get(name))
            expected_signal_type = self._vector_device_type[name] or signal_type
            vector[vector_index] = signal_type(backend)
        elif child := getattr(self._device, name, None):
            # There is an existing child, so raise
            self._raise(name, f"Cannot make child as it would shadow {child}")
        else:
            # We need to add a new child to the top level Device
            backend = self._signal_backend_factory(None)
            expected_signal_type = signal_type
            setattr(self._device, name, signal_type(backend))
        if signal_type is not expected_signal_type:
            self._raise(
                name,
                f"is a {signal_type.__name__} not a {expected_signal_type.__name__}",
            )
        return backend

    def fill_child_device(
        self,
        name: str,
        device_type: type[Device] = Device,
        vector_index: int | None = None,
    ) -> DeviceConnectorT:
        name = cast(LogicalName, name)
        if name in self._unfilled_connectors:
            # We made it above
            connector = self._unfilled_connectors.pop(name)
            self._filled_connectors[name] = connector
        elif name in self._filled_backends:
            # We made it and filled it so return for validation
            connector = self._filled_connectors[name]
        elif vector_index:
            # We need to add a new entry to a DeviceVector
            vector = self._ensure_device_vector(name)
            vector_device_type = self._vector_device_type[name] or device_type
            assert issubclass(
                vector_device_type, Device
            ), f"{vector_device_type} is not a Device"
            connector = self._device_connector_factory()
            vector[vector_index] = vector_device_type(connector=connector)
        elif child := getattr(self._device, name, None):
            # There is an existing child, so raise
            self._raise(name, f"Cannot make child as it would shadow {child}")
        else:
            # We need to add a new child to the top level Device
            connector = self._device_connector_factory()
            setattr(self._device, name, device_type(connector=connector))
        return connector



from ophyd_async.core import StrictEnum
from ophyd_async.epics import adcore
from ophyd_async.epics.core import epics_signal_rw_rbv


class KinetixTriggerMode(StrictEnum):
    internal = "Internal"
    edge = "Rising Edge"
    gate = "Exp. Gate"


class KinetixReadoutMode(StrictEnum):
    sensitivity = 1
    speed = 2
    dynamic_range = 3
    sub_electron = 4


class KinetixDriverIO(adcore.ADBaseIO):
    """This mirrors the interface provided by ADKinetix/db/ADKinetix.template."""

    def __init__(self, prefix: str, name: str = "") -> None:
        # self.pixel_format = epics_signal_rw_rbv(PixelFormat, prefix + "PixelFormat")
        self.trigger_mode = epics_signal_rw_rbv(
            KinetixTriggerMode, prefix + "TriggerMode"
        )
        self.readout_port_idx = epics_signal_rw_rbv(
            KinetixReadoutMode, prefix + "ReadoutPortIdx"
        )
        super().__init__(prefix, name)



from collections.abc import Sequence

from ophyd_async.core import PathProvider, SignalR, StandardDetector
from ophyd_async.epics import adcore

from ._sim_controller import SimController


class SimDetector(StandardDetector):
    _controller: SimController
    _writer: adcore.ADHDFWriter

    def __init__(
        self,
        prefix: str,
        path_provider: PathProvider,
        drv_suffix="cam1:",
        hdf_suffix="HDF1:",
        name: str = "",
        config_sigs: Sequence[SignalR] = (),
    ):
        self.drv = adcore.ADBaseIO(prefix + drv_suffix)
        self.hdf = adcore.NDFileHDFIO(prefix + hdf_suffix)

        super().__init__(
            SimController(self.drv),
            adcore.ADHDFWriter(
                self.hdf,
                path_provider,
                lambda: self.name,
                adcore.ADBaseDatasetDescriber(self.drv),
            ),
            config_sigs=(self.drv.acquire_period, self.drv.acquire_time, *config_sigs),
            name=name,
        )



import asyncio

from bluesky.protocols import (
    Flyable,
    Locatable,
    Location,
    Preparable,
    Stoppable,
)
from pydantic import BaseModel, Field

from ophyd_async.core import (
    CALCULATE_TIMEOUT,
    DEFAULT_TIMEOUT,
    AsyncStatus,
    CalculatableTimeout,
    StandardReadable,
    WatchableAsyncStatus,
    WatcherUpdate,
    observe_value,
)
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.epics.core import epics_signal_r, epics_signal_rw, epics_signal_x


class MotorLimitsException(Exception):
    pass


class InvalidFlyMotorException(Exception):
    pass


DEFAULT_MOTOR_FLY_TIMEOUT = 60
DEFAULT_WATCHER_UPDATE_FREQUENCY = 0.2


class FlyMotorInfo(BaseModel):
    """Minimal set of information required to fly a motor:"""

    #: Absolute position of the motor once it finishes accelerating to desired
    #: velocity, in motor EGUs
    start_position: float = Field(frozen=True)

    #: Absolute position of the motor once it begins decelerating from desired
    #: velocity, in EGUs
    end_position: float = Field(frozen=True)

    #: Time taken for the motor to get from start_position to end_position, excluding
    #: run-up and run-down, in seconds.
    time_for_move: float = Field(frozen=True, gt=0)

    #: Maximum time for the complete motor move, including run up and run down.
    #: Defaults to `time_for_move` + run up and run down times + 10s.
    timeout: CalculatableTimeout = Field(frozen=True, default=CALCULATE_TIMEOUT)


class Motor(StandardReadable, Locatable, Stoppable, Flyable, Preparable):
    """Device that moves a motor record"""

    def __init__(self, prefix: str, name="") -> None:
        # Define some signals
        with self.add_children_as_readables(Format.CONFIG_SIGNAL):
            self.motor_egu = epics_signal_r(str, prefix + ".EGU")
            self.velocity = epics_signal_rw(float, prefix + ".VELO")

        with self.add_children_as_readables(Format.HINTED_SIGNAL):
            self.user_readback = epics_signal_r(float, prefix + ".RBV")

        self.user_setpoint = epics_signal_rw(float, prefix + ".VAL")
        self.max_velocity = epics_signal_r(float, prefix + ".VMAX")
        self.acceleration_time = epics_signal_rw(float, prefix + ".ACCL")
        self.precision = epics_signal_r(int, prefix + ".PREC")
        self.deadband = epics_signal_r(float, prefix + ".RDBD")
        self.motor_done_move = epics_signal_r(int, prefix + ".DMOV")
        self.low_limit_travel = epics_signal_rw(float, prefix + ".LLM")
        self.high_limit_travel = epics_signal_rw(float, prefix + ".HLM")

        self.motor_stop = epics_signal_x(prefix + ".STOP")
        # Whether set() should complete successfully or not
        self._set_success = True

        # end_position of a fly move, with run_up_distance added on.
        self._fly_completed_position: float | None = None

        # Set on kickoff(), complete when motor reaches self._fly_completed_position
        self._fly_status: WatchableAsyncStatus | None = None

        # Set during prepare
        self._fly_timeout: CalculatableTimeout | None = CALCULATE_TIMEOUT

        super().__init__(name=name)

    def set_name(self, name: str):
        super().set_name(name)
        # Readback should be named the same as its parent in read()
        self.user_readback.set_name(name)

    @AsyncStatus.wrap
    async def prepare(self, value: FlyMotorInfo):
        """Calculate required velocity and run-up distance, then if motor limits aren't
        breached, move to start position minus run-up distance"""

        self._fly_timeout = value.timeout

        # Velocity, at which motor travels from start_position to end_position, in motor
        # egu/s.
        fly_velocity = await self._prepare_velocity(
            value.start_position,
            value.end_position,
            value.time_for_move,
        )

        # start_position with run_up_distance added on.
        fly_prepared_position = await self._prepare_motor_path(
            abs(fly_velocity), value.start_position, value.end_position
        )

        await self.set(fly_prepared_position)
        await self.velocity.set(abs(fly_velocity))

    @AsyncStatus.wrap
    async def kickoff(self):
        """Begin moving motor from prepared position to final position."""
        assert (
            self._fly_completed_position
        ), "Motor must be prepared before attempting to kickoff"

        self._fly_status = self.set(
            self._fly_completed_position, timeout=self._fly_timeout
        )

    def complete(self) -> WatchableAsyncStatus:
        """Mark as complete once motor reaches completed position."""
        assert self._fly_status, "kickoff not called"
        return self._fly_status

    @WatchableAsyncStatus.wrap
    async def set(self, value: float, timeout: CalculatableTimeout = CALCULATE_TIMEOUT):
        new_position = value
        self._set_success = True
        (
            old_position,
            units,
            precision,
            velocity,
            acceleration_time,
        ) = await asyncio.gather(
            self.user_setpoint.get_value(),
            self.motor_egu.get_value(),
            self.precision.get_value(),
            self.velocity.get_value(),
            self.acceleration_time.get_value(),
        )
        if timeout is CALCULATE_TIMEOUT:
            assert velocity > 0, "Motor has zero velocity"
            timeout = (
                abs(new_position - old_position) / velocity
                + 2 * acceleration_time
                + DEFAULT_TIMEOUT
            )
        move_status = self.user_setpoint.set(new_position, wait=True, timeout=timeout)
        async for current_position in observe_value(
            self.user_readback, done_status=move_status
        ):
            yield WatcherUpdate(
                current=current_position,
                initial=old_position,
                target=new_position,
                name=self.name,
                unit=units,
                precision=precision,
            )
        if not self._set_success:
            raise RuntimeError("Motor was stopped")

    async def stop(self, success=False):
        self._set_success = success
        # Put with completion will never complete as we are waiting for completion on
        # the move above, so need to pass wait=False
        await self.motor_stop.trigger(wait=False)

    async def _prepare_velocity(
        self, start_position: float, end_position: float, time_for_move: float
    ) -> float:
        fly_velocity = (end_position - start_position) / time_for_move
        max_speed, egu = await asyncio.gather(
            self.max_velocity.get_value(), self.motor_egu.get_value()
        )
        if abs(fly_velocity) > max_speed:
            raise MotorLimitsException(
                f"Motor speed of {abs(fly_velocity)} {egu}/s was requested for a motor "
                f" with max speed of {max_speed} {egu}/s"
            )
        # move to prepare position at maximum velocity
        await self.velocity.set(abs(max_speed))
        return fly_velocity

    async def locate(self) -> Location[float]:
        location: Location = {
            "setpoint": await self.user_setpoint.get_value(),
            "readback": await self.user_readback.get_value(),
        }
        return location

    async def _prepare_motor_path(
        self, fly_velocity: float, start_position: float, end_position: float
    ) -> float:
        # Distance required for motor to accelerate from stationary to fly_velocity, and
        # distance required for motor to decelerate from fly_velocity to stationary
        run_up_distance = (
            (await self.acceleration_time.get_value()) * fly_velocity * 0.5
        )

        self._fly_completed_position = end_position + run_up_distance

        # Prepared position not used after prepare, so no need to store in self
        fly_prepared_position = start_position - run_up_distance

        motor_lower_limit, motor_upper_limit, egu = await asyncio.gather(
            self.low_limit_travel.get_value(),
            self.high_limit_travel.get_value(),
            self.motor_egu.get_value(),
        )

        if (
            not motor_upper_limit >= fly_prepared_position >= motor_lower_limit
            or not motor_upper_limit
            >= self._fly_completed_position
            >= motor_lower_limit
        ):
            raise MotorLimitsException(
                f"Motor trajectory for requested fly is from "
                f"{fly_prepared_position}{egu} to "
                f"{self._fly_completed_position}{egu} but motor limits are "
                f"{motor_lower_limit}{egu} <= x <= {motor_upper_limit}{egu} "
            )
        return fly_prepared_position



from ophyd_async.core import (
    Device,
    DeviceVector,
    SignalR,
    SignalRW,
    StrictEnum,
    SubsetEnum,
)

from ._table import DatasetTable, SeqTable


class CaptureMode(StrictEnum):
    FIRST_N = "FIRST_N"
    LAST_N = "LAST_N"
    FOREVER = "FOREVER"


class DataBlock(Device):
    # In future we may decide to make hdf_* optional
    hdf_directory: SignalRW[str]
    hdf_file_name: SignalRW[str]
    num_capture: SignalRW[int]
    num_captured: SignalR[int]
    create_directory: SignalRW[int]
    directory_exists: SignalR[bool]
    capture_mode: SignalRW[CaptureMode]
    capture: SignalRW[bool]
    flush_period: SignalRW[float]
    datasets: SignalR[DatasetTable]


class PulseBlock(Device):
    delay: SignalRW[float]
    width: SignalRW[float]


class PcompDirection(StrictEnum):
    positive = "Positive"
    negative = "Negative"
    either = "Either"


class BitMux(SubsetEnum):
    zero = "ZERO"
    one = "ONE"


class PcompBlock(Device):
    active: SignalR[bool]
    dir: SignalRW[PcompDirection]
    enable: SignalRW[BitMux]
    pulses: SignalRW[int]
    start: SignalRW[int]
    step: SignalRW[int]
    width: SignalRW[int]


class TimeUnits(StrictEnum):
    min = "min"
    s = "s"
    ms = "ms"
    us = "us"


class SeqBlock(Device):
    table: SignalRW[SeqTable]
    active: SignalR[bool]
    repeats: SignalRW[int]
    prescale: SignalRW[float]
    prescale_units: SignalRW[TimeUnits]
    enable: SignalRW[BitMux]


class PcapBlock(Device):
    active: SignalR[bool]
    arm: SignalRW[bool]


class CommonPandaBlocks(Device):
    pulse: DeviceVector[PulseBlock]
    seq: DeviceVector[SeqBlock]
    pcomp: DeviceVector[PcompBlock]
    pcap: PcapBlock
    data: DataBlock



from pathlib import Path

import pytest

from ophyd_async.sim.demo import PatternDetector


@pytest.fixture
async def sim_pattern_detector(tmp_path_factory) -> PatternDetector:
    path: Path = tmp_path_factory.mktemp("tmp")
    return PatternDetector(name="PATTERN1", path=path)



import asyncio

import numpy as np
from bluesky.protocols import Movable, Stoppable

from ophyd_async.core import (
    CALCULATE_TIMEOUT,
    DEFAULT_TIMEOUT,
    AsyncStatus,
    CalculatableTimeout,
    Device,
    StandardReadable,
    WatchableAsyncStatus,
    WatcherUpdate,
    observe_value,
)
from ophyd_async.core import StandardReadableFormat as Format
from ophyd_async.epics.core import epics_signal_r, epics_signal_rw, epics_signal_x


class Mover(StandardReadable, Movable, Stoppable):
    """A demo movable that moves based on velocity"""

    def __init__(self, prefix: str, name="") -> None:
        # Define some signals
        with self.add_children_as_readables(Format.HINTED_SIGNAL):
            self.readback = epics_signal_r(float, prefix + "Readback")
        with self.add_children_as_readables(Format.CONFIG_SIGNAL):
            self.velocity = epics_signal_rw(float, prefix + "Velocity")
            self.units = epics_signal_r(str, prefix + "Readback.EGU")
        self.setpoint = epics_signal_rw(float, prefix + "Setpoint")
        self.precision = epics_signal_r(int, prefix + "Readback.PREC")
        # Signals that collide with standard methods should have a trailing underscore
        self.stop_ = epics_signal_x(prefix + "Stop.PROC")
        # Whether set() should complete successfully or not
        self._set_success = True

        super().__init__(name=name)

    def set_name(self, name: str):
        super().set_name(name)
        # Readback should be named the same as its parent in read()
        self.readback.set_name(name)

    @WatchableAsyncStatus.wrap
    async def set(self, value: float, timeout: CalculatableTimeout = CALCULATE_TIMEOUT):
        new_position = value
        self._set_success = True
        old_position, units, precision, velocity = await asyncio.gather(
            self.setpoint.get_value(),
            self.units.get_value(),
            self.precision.get_value(),
            self.velocity.get_value(),
        )
        if timeout == CALCULATE_TIMEOUT:
            assert velocity > 0, "Mover has zero velocity"
            timeout = abs(new_position - old_position) / velocity + DEFAULT_TIMEOUT
        # Make an Event that will be set on completion, and a Status that will
        # error if not done in time
        done = asyncio.Event()
        done_status = AsyncStatus(asyncio.wait_for(done.wait(), timeout))
        # Wait for the value to set, but don't wait for put completion callback
        await self.setpoint.set(new_position, wait=False)
        async for current_position in observe_value(
            self.readback, done_status=done_status
        ):
            yield WatcherUpdate(
                current=current_position,
                initial=old_position,
                target=new_position,
                name=self.name,
                unit=units,
                precision=precision,
            )
            if np.isclose(current_position, new_position):
                done.set()
                break
        if not self._set_success:
            raise RuntimeError("Motor was stopped")

    async def stop(self, success=True):
        self._set_success = success
        status = self.stop_.trigger()
        await status


class SampleStage(Device):
    """A demo sample stage with X and Y movables"""

    def __init__(self, prefix: str, name="") -> None:
        # Define some child Devices
        self.x = Mover(prefix + "X:")
        self.y = Mover(prefix + "Y:")
        # Set name of device and child devices
        super().__init__(name=name)



import asyncio
import subprocess
from collections import defaultdict
from unittest.mock import ANY, Mock, call, patch

import pytest
from bluesky import plans as bp
from bluesky.protocols import Reading
from bluesky.run_engine import RunEngine

from ophyd_async.core import (
    DeviceCollector,
    NotConnected,
    assert_emitted,
    assert_reading,
    assert_value,
    callback_on_mock_put,
    get_mock,
    get_mock_put,
    set_mock_value,
)
from ophyd_async.epics import demo

# Long enough for multiple asyncio event loop cycles to run so
# all the tasks have a chance to run
A_WHILE = 0.001


@pytest.fixture
async def mock_mover() -> demo.Mover:
    async with DeviceCollector(mock=True):
        mock_mover = demo.Mover("BLxxI-MO-TABLE-01:X:")
        # Signals connected here

    assert mock_mover.name == "mock_mover"
    set_mock_value(mock_mover.units, "mm")
    set_mock_value(mock_mover.precision, 3)
    set_mock_value(mock_mover.velocity, 1)
    return mock_mover


@pytest.fixture
async def mock_sensor() -> demo.Sensor:
    async with DeviceCollector(mock=True):
        mock_sensor = demo.Sensor("MOCK:SENSOR:")
        # Signals connected here

    assert mock_sensor.name == "mock_sensor"
    return mock_sensor


@pytest.fixture
async def mock_sensor_group() -> demo.SensorGroup:
    async with DeviceCollector(mock=True):
        mock_sensor_group = demo.SensorGroup("MOCK:SENSOR:")
        # Signals connected here

    assert mock_sensor_group.name == "mock_sensor_group"
    return mock_sensor_group


async def test_mover_stopped(mock_mover: demo.Mover):
    callbacks = []
    callback_on_mock_put(
        mock_mover.stop_, lambda v, *args, **kwargs: callbacks.append(v)
    )

    await mock_mover.stop()
    assert callbacks == [None]


class DemoWatcher:
    def __init__(self) -> None:
        self._event = asyncio.Event()
        self._mock = Mock()

    def __call__(
        self,
        *args,
        current: float,
        initial: float,
        target: float,
        name: str | None = None,
        unit: str | None = None,
        precision: float | None = None,
        fraction: float | None = None,
        time_elapsed: float | None = None,
        time_remaining: float | None = None,
        **kwargs,
    ):
        self._mock(
            *args,
            current=current,
            initial=initial,
            target=target,
            name=name,
            unit=unit,
            precision=precision,
            time_elapsed=time_elapsed,
            **kwargs,
        )
        self._event.set()

    async def wait_for_call(self, *args, **kwargs):
        await asyncio.wait_for(self._event.wait(), timeout=1)
        assert self._mock.call_count == 1
        assert self._mock.call_args == call(*args, **kwargs)
        self._mock.reset_mock()
        self._event.clear()


async def test_mover_moving_well(mock_mover: demo.Mover) -> None:
    s = mock_mover.set(0.55)
    watcher = DemoWatcher()
    s.watch(watcher)
    done = Mock()
    s.add_callback(done)
    await watcher.wait_for_call(
        name="mock_mover",
        current=0.0,
        initial=0.0,
        target=0.55,
        unit="mm",
        precision=3,
        time_elapsed=pytest.approx(0.0, abs=0.05),
    )

    await assert_value(mock_mover.setpoint, 0.55)
    assert not s.done
    done.assert_not_called()
    await asyncio.sleep(0.1)
    set_mock_value(mock_mover.readback, 0.1)
    await watcher.wait_for_call(
        name="mock_mover",
        current=0.1,
        initial=0.0,
        target=0.55,
        unit="mm",
        precision=3,
        time_elapsed=pytest.approx(0.1, abs=0.05),
    )
    set_mock_value(mock_mover.readback, 0.5499999)
    await asyncio.sleep(A_WHILE)
    assert s.done
    assert s.success
    done.assert_called_once_with(s)
    done2 = Mock()
    s.add_callback(done2)
    done2.assert_called_once_with(s)


async def test_sensor_reading_shows_value(mock_sensor: demo.Sensor):
    # Check default value
    await assert_value(mock_sensor.value, pytest.approx(0.0))
    assert (await mock_sensor.value.get_value()) == pytest.approx(0.0)
    await assert_reading(
        mock_sensor,
        {
            "mock_sensor-value": {
                "value": 0.0,
                "alarm_severity": 0,
                "timestamp": ANY,
            }
        },
    )
    # Check different value
    set_mock_value(mock_sensor.value, 5.0)
    await assert_reading(
        mock_sensor,
        {
            "mock_sensor-value": {
                "value": 5.0,
                "timestamp": ANY,
                "alarm_severity": 0,
            }
        },
    )


async def test_retrieve_mock_and_assert(mock_mover: demo.Mover):
    mover_setpoint_mock = get_mock_put(mock_mover.setpoint)
    await mock_mover.setpoint.set(10)
    mover_setpoint_mock.assert_called_once_with(10, wait=ANY)

    # Assert that velocity is set before move
    mover_velocity_mock = get_mock_put(mock_mover.velocity)

    parent_mock = Mock()
    parent_mock.attach_mock(mover_setpoint_mock, "setpoint")
    parent_mock.attach_mock(mover_velocity_mock, "velocity")

    await mock_mover.velocity.set(100)
    await mock_mover.setpoint.set(67)
    assert parent_mock.mock_calls == [
        call.velocity(100, wait=True),
        call.setpoint(67, wait=True),
    ]


async def test_mocks_in_device_share_parent():
    mock = Mock()
    async with DeviceCollector(mock=mock):
        mock_mover = demo.Mover("BLxxI-MO-TABLE-01:Y:")

    assert get_mock(mock_mover) is mock
    assert get_mock(mock_mover.setpoint) is mock.setpoint
    assert get_mock_put(mock_mover.setpoint) is mock.setpoint.put
    await mock_mover.setpoint.set(10)
    get_mock_put(mock_mover.setpoint).assert_called_once_with(10, wait=ANY)

    await mock_mover.velocity.set(100)
    await mock_mover.setpoint.set(67)

    mock.reset_mock()
    await mock_mover.velocity.set(100)
    await mock_mover.setpoint.set(67)
    assert mock.mock_calls == [
        call.velocity.put(100, wait=True),
        call.setpoint.put(67, wait=True),
    ]


async def test_read_mover(mock_mover: demo.Mover):
    await mock_mover.stage()
    assert (await mock_mover.read())["mock_mover"]["value"] == 0.0
    assert (await mock_mover.read_configuration())["mock_mover-velocity"]["value"] == 1
    assert (await mock_mover.describe_configuration())["mock_mover-units"][
        "shape"
    ] == []
    set_mock_value(mock_mover.readback, 0.5)
    assert (await mock_mover.read())["mock_mover"]["value"] == 0.5
    await mock_mover.unstage()
    # Check we can still read and describe when not staged
    set_mock_value(mock_mover.readback, 0.1)
    assert (await mock_mover.read())["mock_mover"]["value"] == 0.1
    assert await mock_mover.describe()


async def test_set_velocity(mock_mover: demo.Mover) -> None:
    v = mock_mover.velocity
    q: asyncio.Queue[dict[str, Reading]] = asyncio.Queue()
    v.subscribe(q.put_nowait)
    assert (await q.get())["mock_mover-velocity"]["value"] == 1.0
    await v.set(2.0)
    assert (await q.get())["mock_mover-velocity"]["value"] == 2.0
    v.clear_sub(q.put_nowait)
    await v.set(3.0)
    assert (await v.read())["mock_mover-velocity"]["value"] == 3.0
    assert q.empty()


async def test_mover_disconnected():
    with pytest.raises(NotConnected):
        async with DeviceCollector(timeout=0.1):
            m = demo.Mover("ca://PRE:", name="mover")
    assert m.name == "mover"


async def test_sensor_disconnected(caplog):
    caplog.set_level(10)
    with pytest.raises(NotConnected):
        async with DeviceCollector(timeout=0.1):
            s = demo.Sensor("ca://PRE:", name="sensor")
    logs = caplog.get_records("call")
    logs = [log for log in logs if "_signal" not in log.pathname]
    assert len(logs) == 2
    messages = {log.message for log in logs}

    assert messages == {
        "signal ca://PRE:Value timed out",
        "signal ca://PRE:Mode timed out",
    }
    assert s.name == "sensor"


async def test_read_sensor(mock_sensor: demo.Sensor):
    assert (await mock_sensor.read())["mock_sensor-value"]["value"] == 0
    assert (await mock_sensor.read_configuration())["mock_sensor-mode"][
        "value"
    ] == demo.EnergyMode.low
    desc = (await mock_sensor.describe_configuration())["mock_sensor-mode"]
    assert desc["dtype"] == "string"
    assert desc["choices"] == ["Low Energy", "High Energy"]
    set_mock_value(mock_sensor.mode, demo.EnergyMode.high)
    assert (await mock_sensor.read_configuration())["mock_sensor-mode"][
        "value"
    ] == demo.EnergyMode.high


async def test_sensor_in_plan(RE: RunEngine, mock_sensor: demo.Sensor):
    """Tests mock sensor behavior within a RunEngine plan.

    This test verifies that the sensor emits the expected documents
     when used in plan(count).
    """
    docs = defaultdict(list)

    def capture_emitted(name, doc):
        docs[name].append(doc)

    RE(bp.count([mock_sensor], num=2), capture_emitted)
    assert_emitted(docs, start=1, descriptor=1, event=2, stop=1)


async def test_assembly_renaming() -> None:
    thing = demo.SampleStage("PRE")
    await thing.connect(mock=True)
    assert thing.x.name == ""
    assert thing.x.velocity.name == ""
    assert thing.x.stop_.name == ""
    await thing.x.velocity.set(456)
    assert await thing.x.velocity.get_value() == 456
    thing.set_name("foo")
    assert thing.x.name == "foo-x"
    assert thing.x.velocity.name == "foo-x-velocity"
    assert thing.x.stop_.name == "foo-x-stop"


async def test_dynamic_sensor_group_disconnected():
    with pytest.raises(NotConnected) as e:
        async with DeviceCollector(timeout=0.1):
            mock_sensor_group_dynamic = demo.SensorGroup("MOCK:SENSOR:")
    expected = """
mock_sensor_group_dynamic: NotConnected:
    sensors: NotConnected:
        1: NotConnected:
            value: NotConnected: ca://MOCK:SENSOR:1:Value
            mode: NotConnected: ca://MOCK:SENSOR:1:Mode
        2: NotConnected:
            value: NotConnected: ca://MOCK:SENSOR:2:Value
            mode: NotConnected: ca://MOCK:SENSOR:2:Mode
        3: NotConnected:
            value: NotConnected: ca://MOCK:SENSOR:3:Value
            mode: NotConnected: ca://MOCK:SENSOR:3:Mode
"""
    assert str(e.value) == expected

    assert mock_sensor_group_dynamic.name == "mock_sensor_group_dynamic"


async def test_dynamic_sensor_group_read_and_describe(
    mock_sensor_group: demo.SensorGroup,
):
    set_mock_value(mock_sensor_group.sensors[1].value, 0.0)
    set_mock_value(mock_sensor_group.sensors[2].value, 0.5)
    set_mock_value(mock_sensor_group.sensors[3].value, 1.0)

    await mock_sensor_group.stage()
    description = await mock_sensor_group.describe()

    await mock_sensor_group.unstage()
    await assert_reading(
        mock_sensor_group,
        {
            "mock_sensor_group-sensors-1-value": {
                "value": 0.0,
                "timestamp": ANY,
                "alarm_severity": 0,
            },
            "mock_sensor_group-sensors-2-value": {
                "value": 0.5,
                "timestamp": ANY,
                "alarm_severity": 0,
            },
            "mock_sensor_group-sensors-3-value": {
                "value": 1.0,
                "timestamp": ANY,
                "alarm_severity": 0,
            },
        },
    )
    assert description == {
        "mock_sensor_group-sensors-1-value": {
            "dtype": "number",
            "dtype_numpy": "<f8",
            "shape": [],
            "source": "mock+ca://MOCK:SENSOR:1:Value",
        },
        "mock_sensor_group-sensors-2-value": {
            "dtype": "number",
            "dtype_numpy": "<f8",
            "shape": [],
            "source": "mock+ca://MOCK:SENSOR:2:Value",
        },
        "mock_sensor_group-sensors-3-value": {
            "dtype": "number",
            "dtype_numpy": "<f8",
            "shape": [],
            "source": "mock+ca://MOCK:SENSOR:3:Value",
        },
    }


@patch("ophyd_async.epics.demo.subprocess.Popen")
async def test_ioc_starts(mock_popen: Mock):
    demo.start_ioc_subprocess()
    mock_popen.assert_called_once_with(
        ANY,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    )



import pytest

from ophyd_async.core import (
    DetectorTrigger,
    StaticPathProvider,
    TriggerInfo,
    set_mock_value,
)
from ophyd_async.epics import adkinetix


@pytest.fixture
def test_adkinetix(ad_standard_det_factory):
    return ad_standard_det_factory(adkinetix.KinetixDetector)


async def test_get_deadtime(
    test_adkinetix: adkinetix.KinetixDetector,
):
    # Currently Kinetix driver doesn't support getting deadtime.
    assert test_adkinetix._controller.get_deadtime(0) == 0.001


async def test_trigger_modes(test_adkinetix: adkinetix.KinetixDetector):
    set_mock_value(
        test_adkinetix.drv.trigger_mode, adkinetix.KinetixTriggerMode.internal
    )

    async def setup_trigger_mode(trig_mode: DetectorTrigger):
        await test_adkinetix.controller.prepare(
            TriggerInfo(number_of_triggers=1, trigger=trig_mode)
        )
        await test_adkinetix.controller.arm()
        await test_adkinetix.controller.wait_for_idle()
        # Prevent timeouts
        set_mock_value(test_adkinetix.drv.acquire, True)

    # Default TriggerSource
    assert (await test_adkinetix.drv.trigger_mode.get_value()) == "Internal"

    await setup_trigger_mode(DetectorTrigger.edge_trigger)
    assert (await test_adkinetix.drv.trigger_mode.get_value()) == "Rising Edge"

    await setup_trigger_mode(DetectorTrigger.constant_gate)
    assert (await test_adkinetix.drv.trigger_mode.get_value()) == "Exp. Gate"

    await setup_trigger_mode(DetectorTrigger.internal)
    assert (await test_adkinetix.drv.trigger_mode.get_value()) == "Internal"

    await setup_trigger_mode(DetectorTrigger.variable_gate)
    assert (await test_adkinetix.drv.trigger_mode.get_value()) == "Exp. Gate"


async def test_hints_from_hdf_writer(test_adkinetix: adkinetix.KinetixDetector):
    assert test_adkinetix.hints == {"fields": [test_adkinetix.name]}


async def test_can_read(test_adkinetix: adkinetix.KinetixDetector):
    # Standard detector can be used as Readable
    assert (await test_adkinetix.read()) == {}


async def test_decribe_describes_writer_dataset(
    test_adkinetix: adkinetix.KinetixDetector, one_shot_trigger_info: TriggerInfo
):
    assert await test_adkinetix.describe() == {}
    await test_adkinetix.stage()
    await test_adkinetix.prepare(one_shot_trigger_info)
    assert await test_adkinetix.describe() == {
        "test_adkinetix1": {
            "source": "mock+ca://KINETIX1:HDF1:FullFileName_RBV",
            "shape": [10, 10],
            "dtype": "array",
            "dtype_numpy": "|i1",
            "external": "STREAM:",
        }
    }


async def test_can_collect(
    test_adkinetix: adkinetix.KinetixDetector,
    static_path_provider: StaticPathProvider,
    one_shot_trigger_info: TriggerInfo,
):
    path_info = static_path_provider()
    full_file_name = path_info.directory_path / f"{path_info.filename}.h5"

    await test_adkinetix.stage()
    await test_adkinetix.prepare(one_shot_trigger_info)
    docs = [(name, doc) async for name, doc in test_adkinetix.collect_asset_docs(1)]
    assert len(docs) == 2
    assert docs[0][0] == "stream_resource"
    stream_resource = docs[0][1]
    sr_uid = stream_resource["uid"]
    assert stream_resource["data_key"] == "test_adkinetix1"
    assert stream_resource["uri"] == "file://localhost/" + str(full_file_name).lstrip(
        "/"
    )
    assert stream_resource["parameters"] == {
        "dataset": "/entry/data/data",
        "swmr": False,
        "multiplier": 1,
        "chunk_shape": (1, 10, 10),
    }
    assert docs[1][0] == "stream_datum"
    stream_datum = docs[1][1]
    assert stream_datum["stream_resource"] == sr_uid
    assert stream_datum["seq_nums"] == {"start": 0, "stop": 0}
    assert stream_datum["indices"] == {"start": 0, "stop": 1}


async def test_can_decribe_collect(
    test_adkinetix: adkinetix.KinetixDetector, one_shot_trigger_info: TriggerInfo
):
    assert (await test_adkinetix.describe_collect()) == {}
    await test_adkinetix.stage()
    await test_adkinetix.prepare(one_shot_trigger_info)
    assert (await test_adkinetix.describe_collect()) == {
        "test_adkinetix1": {
            "source": "mock+ca://KINETIX1:HDF1:FullFileName_RBV",
            "shape": [10, 10],
            "dtype": "array",
            "dtype_numpy": "|i1",
            "external": "STREAM:",
        }
    }



from ._pattern_detector import (
    DATA_PATH,
    SUM_PATH,
    PatternDetector,
    PatternDetectorController,
    PatternDetectorWriter,
    PatternGenerator,
)
from ._sim_motor import SimMotor

__all__ = [
    "DATA_PATH",
    "SUM_PATH",
    "PatternGenerator",
    "PatternDetector",
    "PatternDetectorController",
    "PatternDetectorWriter",
    "SimMotor",
]



"""Demo EPICS Devices for the tutorial"""

import atexit
import random
import string
import subprocess
import sys
from pathlib import Path

from ._mover import Mover, SampleStage
from ._sensor import EnergyMode, Sensor, SensorGroup

__all__ = [
    "Mover",
    "SampleStage",
    "EnergyMode",
    "Sensor",
    "SensorGroup",
]


def start_ioc_subprocess() -> str:
    """Start an IOC subprocess with EPICS database for sample stage and sensor
    with the same pv prefix
    """

    pv_prefix = "".join(random.choice(string.ascii_uppercase) for _ in range(12)) + ":"
    here = Path(__file__).absolute().parent
    args = [sys.executable, "-m", "epicscorelibs.ioc"]

    # Create standalone sensor
    args += ["-m", f"P={pv_prefix}"]
    args += ["-d", str(here / "sensor.db")]

    # Create sensor group
    for suffix in ["1", "2", "3"]:
        args += ["-m", f"P={pv_prefix}{suffix}:"]
        args += ["-d", str(here / "sensor.db")]

    # Create X and Y motors
    for suffix in ["X", "Y"]:
        args += ["-m", f"P={pv_prefix}{suffix}:"]
        args += ["-d", str(here / "mover.db")]

    # Start IOC
    process = subprocess.Popen(
        args,
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    )
    atexit.register(process.communicate, "exit")
    return pv_prefix



import os
from collections import defaultdict

import bluesky.plans as bp
import h5py
import numpy as np
from bluesky import RunEngine

from ophyd_async.core import assert_emitted
from ophyd_async.plan_stubs import ensure_connected
from ophyd_async.sim.demo import PatternDetector


async def test_sim_pattern_detector_initialization(
    sim_pattern_detector: PatternDetector,
):
    assert (
        sim_pattern_detector.pattern_generator
    ), "PatternGenerator was not initialized correctly."


async def test_detector_creates_controller_and_writer(
    sim_pattern_detector: PatternDetector,
):
    assert sim_pattern_detector.writer
    assert sim_pattern_detector.controller


def test_writes_pattern_to_file(
    sim_pattern_detector: PatternDetector,
    RE: RunEngine,
):
    # assert that the file contains data in expected dimensions
    docs = defaultdict(list)

    def capture_emitted(name, doc):
        docs[name].append(doc)

    def plan():
        yield from ensure_connected(sim_pattern_detector, mock=True)
        yield from bp.count([sim_pattern_detector])

    RE(plan(), capture_emitted)
    assert_emitted(
        docs, start=1, descriptor=1, stream_resource=2, stream_datum=2, event=1, stop=1
    )
    path = docs["stream_resource"][0]["uri"].split("://localhost")[-1]
    if os.name == "nt":
        path = path.lstrip("/")
    h5file = h5py.File(path)
    assert list(h5file["/entry"]) == ["data", "sum"]
    assert list(h5file["/entry/sum"]) == [44540.0]
    assert np.sum(h5file["/entry/data/data"]) == 44540.0



import asyncio
import os
import random
import string
import subprocess
import sys
import time
from collections.abc import Sequence
from contextlib import closing
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from types import GenericAlias
from typing import Any, Literal
from unittest.mock import ANY

import bluesky.plan_stubs as bps
import numpy as np
import pytest
from aioca import purge_channel_caches
from bluesky.protocols import Reading
from bluesky.run_engine import RunEngine
from event_model import DataKey, Limits, LimitsRange
from ophyd.signal import EpicsSignal

from ophyd_async.core import (
    Array1D,
    NotConnected,
    SignalBackend,
    SignalRW,
    StrictEnum,
    SubsetEnum,
    T,
    Table,
    load_from_yaml,
    save_to_yaml,
)
from ophyd_async.epics.core import (
    epics_signal_r,
    epics_signal_rw,
    epics_signal_rw_rbv,
    epics_signal_w,
    epics_signal_x,
)
from ophyd_async.epics.core._signal import _epics_signal_backend  # noqa: PLC2701

RECORDS = str(Path(__file__).parent / "test_records.db")
PV_PREFIX = "".join(random.choice(string.ascii_lowercase) for _ in range(12))


@dataclass
class IOC:
    process: subprocess.Popen
    protocol: Literal["ca", "pva"]

    async def make_backend(
        self, typ: type | None, suff: str, timeout=10.0
    ) -> SignalBackend:
        # Calculate the pv
        pv = f"{self.protocol}://{PV_PREFIX}:{self.protocol}:{suff}"
        # Make and connect the backend
        backend = _epics_signal_backend(typ, pv, pv)
        await backend.connect(timeout=timeout)
        return backend


# Use a module level fixture per protocol so it's fast to run tests. This means
# we need to add a record for every PV that we will modify in tests to stop
# tests interfering with each other
@pytest.fixture(scope="module", params=["pva", "ca"])
def ioc(request: pytest.FixtureRequest):
    protocol = request.param
    process = subprocess.Popen(
        [
            sys.executable,
            "-m",
            "epicscorelibs.ioc",
            "-m",
            f"P={PV_PREFIX}:{protocol}:",
            "-d",
            RECORDS,
        ],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    )

    start_time = time.monotonic()
    line = ""
    while "iocRun: All initialization complete" not in line:
        if line:
            print(line)
        if time.monotonic() - start_time > 10:
            raise TimeoutError("IOC did not start in time")
        line = process.stdout.readline().strip()  # type: ignore

    yield IOC(process, protocol)

    # close backend caches before the event loop
    purge_channel_caches()
    try:
        print(process.communicate("exit()")[0])
    except ValueError:
        # Someone else already called communicate
        pass


def assert_types_are_equal(t_actual, t_expected, actual_value):
    expected_plain_type = getattr(t_expected, "__origin__", t_expected)
    if issubclass(expected_plain_type, np.ndarray):
        actual_plain_type = getattr(t_actual, "__origin__", t_actual)
        assert actual_plain_type == expected_plain_type
        actual_dtype_type = actual_value.dtype.type
        expected_dtype_type = t_expected.__args__[1].__args__[0]
        assert actual_dtype_type == expected_dtype_type
    elif (
        expected_plain_type is not str
        and not issubclass(expected_plain_type, Enum)
        and issubclass(expected_plain_type, Sequence)
    ):
        actual_plain_type = getattr(t_actual, "__origin__", t_actual)
        assert issubclass(actual_plain_type, expected_plain_type)
        assert len(actual_value) == 0 or isinstance(
            actual_value[0], t_expected.__args__[0]
        )
    else:
        assert t_actual == t_expected


class MonitorQueue:
    def __init__(self, backend: SignalBackend):
        self.backend = backend
        self.updates: asyncio.Queue[Reading] = asyncio.Queue()
        self.subscription = backend.set_callback(self.updates.put_nowait)

    async def assert_updates(self, expected_value, expected_type=None):
        expected_reading = {
            "value": expected_value,
            "timestamp": pytest.approx(time.time(), rel=0.1),
            "alarm_severity": 0,
        }
        backend_reading = await asyncio.wait_for(self.backend.get_reading(), timeout=5)
        backend_value = await asyncio.wait_for(self.backend.get_value(), timeout=5)
        update_reading = await asyncio.wait_for(self.updates.get(), timeout=5)
        update_value = update_reading["value"]

        assert update_value == expected_value == backend_value
        if expected_type:
            assert_types_are_equal(type(update_value), expected_type, update_value)
            assert_types_are_equal(type(backend_value), expected_type, backend_value)
        assert update_reading == expected_reading == backend_reading

    def close(self):
        self.backend.set_callback(None)


def _is_numpy_subclass(t):
    if t is None:
        return False
    plain_type = t.__origin__ if isinstance(t, GenericAlias) else t
    return issubclass(plain_type, np.ndarray)


async def assert_monitor_then_put(
    ioc: IOC,
    suffix: str,
    datakey: dict,
    initial_value: T,
    put_value: T,
    datatype: type[T] | None = None,
    check_type: bool | None = True,
):
    backend = await ioc.make_backend(datatype, suffix)
    # Make a monitor queue that will monitor for updates
    q = MonitorQueue(backend)
    try:
        # Check datakey
        source = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:{suffix}"
        assert dict(source=source, **datakey) == await backend.get_datakey(source)
        # Check initial value
        await q.assert_updates(
            pytest.approx(initial_value),
            datatype if check_type else None,
        )
        # Put to new value and check that
        await backend.put(put_value, wait=True)
        await q.assert_updates(
            pytest.approx(put_value), datatype if check_type else None
        )
    finally:
        q.close()


class MyEnum(StrictEnum):
    a = "Aaa"
    b = "Bbb"
    c = "Ccc"


class MySubsetEnum(SubsetEnum):
    a = "Aaa"
    b = "Bbb"
    c = "Ccc"


_metadata: dict[str, dict[str, dict[str, Any]]] = {
    "ca": {
        "boolean": {"units": ANY, "limits": ANY},
        "integer": {"units": ANY, "limits": ANY},
        "number": {"units": ANY, "limits": ANY, "precision": ANY},
        "enum": {},
        "string": {},
    },
    "pva": {
        "boolean": {},
        "integer": {"units": ANY, "precision": ANY, "limits": ANY},
        "number": {"units": ANY, "precision": ANY, "limits": ANY},
        "enum": {},
        "string": {"units": ANY, "precision": ANY, "limits": ANY},
    },
}


def datakey(protocol: str, suffix: str, value=None) -> DataKey:
    def get_internal_dtype(suffix: str) -> str:
        # uint32, [u]int64 backed by DBR_DOUBLE, have precision
        if "float" in suffix or "uint32" in suffix or "int64" in suffix:
            return "number"
        if "int" in suffix:
            return "integer"
        if "bool" in suffix:
            return "boolean"
        if "enum" in suffix:
            return "enum"
        return "string"

    def get_dtype(suffix: str) -> str:
        if suffix.endswith("a"):
            return "array"
        if "enum" in suffix:
            return "string"
        return get_internal_dtype(suffix)

    def get_dtype_numpy(suffix: str) -> str:  # type: ignore
        if "float32" in suffix:
            return "<f4"
        if "float" in suffix or "double" in suffix:
            return "<f8"  # Unless specifically float 32, use float 64
        if "bool" in suffix:
            return "|b1"
        if "int" in suffix:
            int_str = "|" if "8" in suffix else "<"
            int_str += "u" if "uint" in suffix else "i"
            if "8" in suffix:
                int_str += "1"
            elif "16" in suffix:
                int_str += "2"
            elif "32" in suffix:
                int_str += "4"
            elif "64" in suffix:
                int_str += "8"
            else:
                int_str += (
                    "4"
                    if os.name == "nt" and np.version.version.startswith("1.")
                    else "8"
                )
            return int_str
        if "str" in suffix or "enum" in suffix:
            return "|S40"

    dtype = get_dtype(suffix)
    dtype_numpy = get_dtype_numpy(suffix)

    d = {
        "dtype": dtype,
        "dtype_numpy": dtype_numpy,
        "shape": [len(value)] if dtype == "array" else [],  # type: ignore
    }
    if get_internal_dtype(suffix) == "enum":
        if issubclass(type(value), Enum):
            d["choices"] = [e.value for e in type(value)]  # type: ignore
        else:
            d["choices"] = list(value.choices)  # type: ignore

    d.update(_metadata[protocol].get(get_internal_dtype(suffix), {}))

    return d  # type: ignore


ls1 = "a string that is just longer than forty characters"
ls2 = "another string that is just longer than forty characters"


@pytest.mark.parametrize(
    "datatype, suffix, initial_value, put_value, supported_backends",
    [
        # python builtin scalars
        (int, "int", 42, 43, {"ca", "pva"}),
        (float, "float", 3.141, 43.5, {"ca", "pva"}),
        (str, "str", "hello", "goodbye", {"ca", "pva"}),
        (MyEnum, "enum", MyEnum.b, MyEnum.c, {"ca", "pva"}),
        # numpy arrays of numpy types
        (
            Array1D[np.int8],
            "int8a",
            [-128, 127],
            [-8, 3, 44],
            {"pva"},
        ),
        (
            Array1D[np.uint8],
            "uint8a",
            [0, 255],
            [218],
            {"ca", "pva"},
        ),
        (
            Array1D[np.int16],
            "int16a",
            [-32768, 32767],
            [-855],
            {"ca", "pva"},
        ),
        (
            Array1D[np.uint16],
            "uint16a",
            [0, 65535],
            [5666],
            {"pva"},
        ),
        (
            Array1D[np.int32],
            "int32a",
            [-2147483648, 2147483647],
            [-2],
            {"ca", "pva"},
        ),
        (
            Array1D[np.uint32],
            "uint32a",
            [0, 4294967295],
            [1022233],
            {"pva"},
        ),
        (
            Array1D[np.int64],
            "int64a",
            [-2147483649, 2147483648],
            [-3],
            {"pva"},
        ),
        (
            Array1D[np.uint64],
            "uint64a",
            [0, 4294967297],
            [995444],
            {"pva"},
        ),
        (
            Array1D[np.float32],
            "float32a",
            [0.000002, -123.123],
            [1.0],
            {"ca", "pva"},
        ),
        (
            Array1D[np.float64],
            "float64a",
            [0.1, -12345678.123],
            [0.2],
            {"ca", "pva"},
        ),
        (
            Sequence[str],
            "stra",
            ["five", "six", "seven"],
            ["nine", "ten"],
            {"pva", "ca"},
        ),
        # Can't do long strings until https://github.com/epics-base/pva2pva/issues/17
        # (str, "longstr", ls1, ls2),
        # (str, "longstr2.VAL$", ls1, ls2),
    ],
)
async def test_backend_get_put_monitor(
    ioc: IOC,
    datatype: type[T],
    suffix: str,
    initial_value: T,
    put_value: T,
    tmp_path: Path,
    supported_backends: set[str],
):
    # ca can't support all the types
    for backend in supported_backends:
        assert backend in ["ca", "pva"]
    if ioc.protocol not in supported_backends:
        return
    # With the given datatype, check we have the correct initial value and putting
    # works
    await assert_monitor_then_put(
        ioc,
        suffix,
        datakey(ioc.protocol, suffix, initial_value),  # type: ignore
        initial_value,
        put_value,
        datatype,
    )
    # With datatype guessed from CA/PVA, check we can set it back to the initial value
    await assert_monitor_then_put(
        ioc,
        suffix,
        datakey(ioc.protocol, suffix, put_value),  # type: ignore
        put_value,
        initial_value,
        datatype=None,
    )

    yaml_path = tmp_path / "test.yaml"
    save_to_yaml([{"test": put_value}], yaml_path)
    loaded = load_from_yaml(yaml_path)
    assert np.all(loaded[0]["test"] == put_value)


@pytest.mark.parametrize("suffix", ["bool", "bool_unnamed"])
async def test_bool_conversion_of_enum(ioc: IOC, suffix: str, tmp_path: Path) -> None:
    """Booleans are converted to Short Enumerations with values 0,1 as database does
    not support boolean natively.
    The flow of test_backend_get_put_monitor Gets a value with a dtype of None: we
    cannot tell the difference between an enum with 2 members and a boolean, so
    cannot get a DataKey that does not mutate form.
    This test otherwise performs the same.
    """
    # With the given datatype, check we have the correct initial value and putting
    # works
    await assert_monitor_then_put(
        ioc,
        suffix,
        datakey(ioc.protocol, suffix),
        True,
        False,
        bool,
    )
    # With datatype guessed from CA/PVA, check we can set it back to the initial value
    await assert_monitor_then_put(
        ioc,
        suffix,
        datakey(ioc.protocol, suffix, True),
        False,
        True,
        bool,
    )

    yaml_path = tmp_path / "test.yaml"
    save_to_yaml([{"test": False}], yaml_path)
    loaded = load_from_yaml(yaml_path)
    assert np.all(loaded[0]["test"] is False)


async def test_error_raised_on_disconnected_PV(ioc: IOC) -> None:
    if ioc.protocol == "pva":
        expected = "pva://Disconnect"
    elif ioc.protocol == "ca":
        expected = "ca://Disconnect"
    else:
        raise TypeError()
    backend = await ioc.make_backend(bool, "bool")
    signal = SignalRW(backend)
    # The below will work without error
    await signal.set(False)
    # Change the name of write_pv to mock disconnection
    backend.__setattr__("write_pv", "Disconnect")
    with pytest.raises(asyncio.TimeoutError, match=expected):
        await signal.set(True, timeout=0.1)


class BadEnum(StrictEnum):
    a = "Aaa"
    b = "B"
    c = "Ccc"


def test_enum_equality():
    """Check that we are allowed to replace the passed datatype enum from a signal with
    a version generated from the signal with at least all of the same values, but
    possibly more.
    """

    class GeneratedChoices(StrictEnum):
        a = "Aaa"
        b = "B"
        c = "Ccc"

    class ExtendedGeneratedChoices(StrictEnum):
        a = "Aaa"
        b = "B"
        c = "Ccc"
        d = "Ddd"

    for enum_class in (GeneratedChoices, ExtendedGeneratedChoices):
        assert BadEnum.a == enum_class.a
        assert BadEnum.a.value == enum_class.a
        assert BadEnum.a.value == enum_class.a.value
        assert BadEnum(enum_class.a) is BadEnum.a
        assert BadEnum(enum_class.a.value) is BadEnum.a
        assert not BadEnum == enum_class

    # We will always PUT BadEnum by String, and GET GeneratedChoices by index,
    # so shouldn't ever run across this from conversion code, but may occur if
    # casting returned values or passing as enum rather than value.
    with pytest.raises(ValueError):
        BadEnum(ExtendedGeneratedChoices.d)


class EnumNoString(Enum):
    a = "Aaa"


class SubsetEnumWrongChoices(SubsetEnum):
    a = "Aaa"
    b = "B"
    c = "Ccc"


@pytest.mark.parametrize(
    "typ, suff, errors",
    [
        (
            BadEnum,
            "enum",
            (
                "has choices ('Aaa', 'Bbb', 'Ccc')",
                "but <enum 'BadEnum'>",
                "requested ['Aaa', 'B', 'Ccc'] to be strictly equal",
            ),
        ),
        (
            SubsetEnumWrongChoices,
            "enum",
            (
                "has choices ('Aaa', 'Bbb', 'Ccc')",
                "but <enum 'SubsetEnumWrongChoices'>",
                "requested ['Aaa', 'B', 'Ccc'] to be a subset",
            ),
        ),
        (
            int,
            "str",
            ("with inferred datatype str", "cannot be coerced to int"),
        ),
        (
            str,
            "float",
            ("with inferred datatype float", "cannot be coerced to str"),
        ),
        (
            str,
            "stra",
            ("with inferred datatype Sequence[str]", "cannot be coerced to str"),
        ),
        (
            int,
            "uint8a",
            ("with inferred datatype Array1D[np.uint8]", "cannot be coerced to int"),
        ),
        (
            float,
            "enum",
            ("with inferred datatype str", "cannot be coerced to float"),
        ),
        (
            Array1D[np.int32],
            "float64a",
            (
                "with inferred datatype Array1D[np.float64]",
                "cannot be coerced to Array1D[np.int32]",
            ),
        ),
    ],
)
async def test_backend_wrong_type_errors(ioc: IOC, typ, suff, errors):
    with pytest.raises(TypeError) as cm:
        await ioc.make_backend(typ, suff)
    for error in errors:
        assert error in str(cm.value)


async def test_backend_put_enum_string(ioc: IOC) -> None:
    backend = await ioc.make_backend(MyEnum, "enum2")
    # Don't do this in production code, but allow on CLI
    await backend.put("Ccc", wait=True)  # type: ignore
    assert MyEnum.c == await backend.get_value()


async def test_backend_enum_which_doesnt_inherit_string(ioc: IOC) -> None:
    with pytest.raises(TypeError):
        backend = await ioc.make_backend(EnumNoString, "enum2")
        await backend.put("Aaa", wait=True)


async def test_backend_get_setpoint(ioc: IOC) -> None:
    backend = await ioc.make_backend(MyEnum, "enum2")
    await backend.put("Ccc", wait=True)
    assert await backend.get_setpoint() == MyEnum.c


def approx_table(datatype: type[Table], table: Table):
    new_table = datatype(**table.model_dump())
    for k, v in new_table:
        if datatype is Table:
            setattr(new_table, k, pytest.approx(v))
        else:
            object.__setattr__(new_table, k, pytest.approx(v))
    return new_table


class MyTable(Table):
    bool: Array1D[np.bool_]
    int: Array1D[np.int32]
    float: Array1D[np.float64]
    str: Sequence[str]
    enum: Sequence[MyEnum]


async def test_pva_table(ioc: IOC) -> None:
    if ioc.protocol == "ca":
        # CA can't do tables
        return
    initial = MyTable(
        bool=np.array([False, False, True, True], np.bool_),
        int=np.array([1, 8, -9, 32], np.int32),
        float=np.array([1.8, 8.2, -6, 32.9887], np.float64),
        str=["Hello", "World", "Foo", "Bar"],
        enum=[MyEnum.a, MyEnum.b, MyEnum.a, MyEnum.c],
    )
    put = MyTable(
        bool=np.array([True, False], np.bool_),
        int=np.array([-5, 32], np.int32),
        float=np.array([8.5, -6.97], np.float64),
        str=["Hello", "Bat"],
        enum=[MyEnum.c, MyEnum.b],
    )
    # Make and connect the backend
    for t, i, p in [(MyTable, initial, put), (None, put, initial)]:
        backend = await ioc.make_backend(t, "table")
        # Make a monitor queue that will monitor for updates
        q = MonitorQueue(backend)
        try:
            # Check datakey
            dk = await backend.get_datakey("test-source")
            expected_dk = {
                "dtype": "array",
                "shape": [len(i)],
                "source": "test-source",
                "dtype_numpy": [
                    # natively bool fields are uint8, so if we don't provide a Table
                    # subclass to specify bool, that is what we get
                    ("bool", "|b1" if t else "|u1"),
                    ("int", "<i4"),
                    ("float", "<f8"),
                    ("str", "|S40"),
                    ("enum", "|S40"),
                ],
            }
            assert expected_dk == dk
            # Check initial value
            await q.assert_updates(approx_table(t or Table, i))
            # Put to new value and check that
            await backend.put(p, wait=True)
            await q.assert_updates(approx_table(t or Table, p))
        finally:
            q.close()


async def test_pva_ntdarray(ioc: IOC):
    if ioc.protocol == "ca":
        # CA can't do ndarray
        return

    put = np.array([1, 2, 3, 4, 5, 6], dtype=np.int64).reshape((2, 3))
    initial = np.zeros_like(put)

    backend = await ioc.make_backend(np.ndarray, "ntndarray")

    # Backdoor into the "raw" data underlying the NDArray in QSrv
    # not supporting direct writes to NDArray at the moment.
    raw_data_backend = await ioc.make_backend(Array1D[np.int64], "ntndarray:data")

    # Make a monitor queue that will monitor for updates
    for i, p in [(initial, put), (put, initial)]:
        with closing(MonitorQueue(backend)) as q:
            assert {
                "source": "test-source",
                "dtype": "array",
                "dtype_numpy": "<i8",
                "shape": [2, 3],
            } == await backend.get_datakey("test-source")
            # Check initial value
            await q.assert_updates(pytest.approx(i))
            await raw_data_backend.put(p.flatten(), wait=True)
            await q.assert_updates(pytest.approx(p))


async def test_writing_to_ndarray_raises_typeerror(ioc: IOC):
    if ioc.protocol == "ca":
        # CA can't do ndarray
        return

    backend = await ioc.make_backend(np.ndarray, "ntndarray")

    with pytest.raises(TypeError):
        await backend.put(np.zeros((6,), dtype=np.int64), wait=True)


async def test_non_existent_errors(ioc: IOC):
    with pytest.raises(NotConnected):
        await ioc.make_backend(str, "non-existent", timeout=0.1)


def test_make_backend_fails_for_different_transports():
    read_pv = "test"
    write_pv = "pva://test"

    with pytest.raises(TypeError) as err:
        epics_signal_rw(str, read_pv, write_pv)
        assert (
            err.args[0]
            == f"Differing transports: {read_pv} has EpicsTransport.ca,"
            + " {write_pv} has EpicsTransport.pva"
        )


def test_signal_helpers():
    read_write = epics_signal_rw(int, "ReadWrite")
    assert read_write._connector.backend.read_pv == "ReadWrite"
    assert read_write._connector.backend.write_pv == "ReadWrite"

    read_write_rbv_manual = epics_signal_rw(int, "ReadWrite_RBV", "ReadWrite")
    assert read_write_rbv_manual._connector.backend.read_pv == "ReadWrite_RBV"
    assert read_write_rbv_manual._connector.backend.write_pv == "ReadWrite"

    read_write_rbv = epics_signal_rw_rbv(int, "ReadWrite")
    assert read_write_rbv._connector.backend.read_pv == "ReadWrite_RBV"
    assert read_write_rbv._connector.backend.write_pv == "ReadWrite"

    read_write_rbv_suffix = epics_signal_rw_rbv(int, "ReadWrite", read_suffix=":RBV")
    assert read_write_rbv_suffix._connector.backend.read_pv == "ReadWrite:RBV"
    assert read_write_rbv_suffix._connector.backend.write_pv == "ReadWrite"

    read = epics_signal_r(int, "Read")
    assert read._connector.backend.read_pv == "Read"

    write = epics_signal_w(int, "Write")
    assert write._connector.backend.write_pv == "Write"

    execute = epics_signal_x("Execute")
    assert execute._connector.backend.write_pv == "Execute"


async def test_str_enum_returns_enum(ioc: IOC):
    await ioc.make_backend(MyEnum, "enum")
    pv_name = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:enum"

    sig = epics_signal_rw(MyEnum, pv_name)
    await sig.connect()
    val = await sig.get_value()
    assert repr(val) == "<MyEnum.b: 'Bbb'>"
    assert val is MyEnum.b
    assert val == "Bbb"


async def test_str_datatype_in_mbbo(ioc: IOC):
    backend = await ioc.make_backend(MyEnum, "enum")
    pv_name = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:enum"
    sig = epics_signal_rw(str, pv_name)
    datakey = await backend.get_datakey(sig.source)
    assert datakey["choices"] == ["Aaa", "Bbb", "Ccc"]
    await sig.connect()
    description = await sig.describe()
    assert description[""]["choices"] == ["Aaa", "Bbb", "Ccc"]
    val = await sig.get_value()
    assert val == "Bbb"


async def test_runtime_enum_returns_str(ioc: IOC):
    await ioc.make_backend(MySubsetEnum, "enum")
    pv_name = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:enum"
    sig = epics_signal_rw(MySubsetEnum, pv_name)

    await sig.connect()
    val = await sig.get_value()
    assert val == "Bbb"


async def test_signal_returns_units_and_precision(ioc: IOC):
    await ioc.make_backend(float, "float")
    pv_name = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:float"

    sig = epics_signal_rw(float, pv_name)
    await sig.connect()
    datakey = (await sig.describe())[""]
    assert datakey["units"] == "mm"
    assert datakey["precision"] == 1


async def test_signal_not_return_none_units_and_precision(ioc: IOC):
    await ioc.make_backend(str, "str")
    pv_name = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:str"

    sig = epics_signal_rw(str, pv_name)
    await sig.connect()
    datakey = (await sig.describe())[""]
    assert not hasattr(datakey, "units")
    assert not hasattr(datakey, "precision")


async def test_signal_returns_limits(ioc: IOC):
    await ioc.make_backend(int, "int")
    pv_name = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:int"

    expected_limits = Limits(
        # LOW, HIGH
        warning=LimitsRange(low=5.0, high=96.0),
        # DRVL, DRVH
        control=LimitsRange(low=10.0, high=90.0),
        # LOPR, HOPR
        display=LimitsRange(low=0.0, high=100.0),
        # LOLO, HIHI
        alarm=LimitsRange(low=2.0, high=98.0),
    )

    sig = epics_signal_rw(int, pv_name)
    await sig.connect()
    limits = (await sig.describe())[""]["limits"]
    assert limits == expected_limits


async def test_signal_returns_partial_limits(ioc: IOC):
    await ioc.make_backend(int, "partialint")
    pv_name = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:partialint"

    expected_limits = Limits(
        # LOLO, HIHI
        alarm=LimitsRange(low=2.0, high=98.0),
        # DRVL, DRVH
        control=LimitsRange(low=10.0, high=90.0),
        # LOPR, HOPR
        display=LimitsRange(low=0.0, high=100.0),
    )
    if ioc.protocol == "ca":
        # HSV, LSV not set, but still present for CA
        expected_limits["warning"] = LimitsRange(low=0, high=0)

    sig = epics_signal_rw(int, pv_name)
    await sig.connect()
    limits = (await sig.describe())[""]["limits"]
    assert limits == expected_limits


async def test_signal_returns_warning_and_partial_limits(ioc: IOC):
    await ioc.make_backend(int, "lessint")
    pv_name = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:lessint"

    expected_limits = Limits(
        # control = display if DRVL, DRVH not set
        control=LimitsRange(low=0.0, high=100.0),
        # LOPR, HOPR
        display=LimitsRange(low=0.0, high=100.0),
        # LOW, HIGH
        warning=LimitsRange(low=2.0, high=98.0),
    )
    if ioc.protocol == "ca":
        # HSV, LSV not set, but still present for CA
        expected_limits["alarm"] = LimitsRange(low=0, high=0)

    sig = epics_signal_rw(int, pv_name)
    await sig.connect()
    limits = (await sig.describe())[""]["limits"]
    assert limits == expected_limits


async def test_signal_not_return_no_limits(ioc: IOC):
    await ioc.make_backend(MyEnum, "enum")
    pv_name = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:enum"
    sig = epics_signal_rw(MyEnum, pv_name)
    await sig.connect()
    datakey = (await sig.describe())[""]
    assert not hasattr(datakey, "limits")


async def test_signals_created_for_prec_0_float_can_use_int(ioc: IOC):
    pv_name = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:float_prec_0"
    sig = epics_signal_rw(int, pv_name)
    await sig.connect()


async def test_signals_created_for_not_prec_0_float_cannot_use_int(ioc: IOC):
    pv_name = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:float_prec_1"
    sig = epics_signal_rw(int, pv_name)
    with pytest.raises(
        TypeError,
        match=f"{ioc.protocol}:float_prec_1 with inferred datatype float"
        ".* cannot be coerced to int",
    ):
        await sig.connect()


async def test_bool_works_for_mismatching_enums(ioc: IOC):
    pv_name = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:bool"
    sig = epics_signal_rw(bool, pv_name, pv_name + "_unnamed")
    await sig.connect()


@pytest.mark.skipif(os.name == "nt", reason="Hangs on windows for unknown reasons")
async def test_can_read_using_ophyd_async_then_ophyd(ioc: IOC):
    oa_read = f"{ioc.protocol}://{PV_PREFIX}:{ioc.protocol}:float_prec_1"
    ophyd_read = f"{PV_PREFIX}:{ioc.protocol}:float_prec_0"

    ophyd_async_sig = epics_signal_rw(float, oa_read)
    await ophyd_async_sig.connect()
    ophyd_signal = EpicsSignal(ophyd_read)
    ophyd_signal.wait_for_connection(timeout=5)

    RE = RunEngine()

    def my_plan():
        yield from bps.rd(ophyd_async_sig)
        yield from bps.rd(ophyd_signal)

    RE(my_plan())


def test_signal_module_emits_deprecation_warning():
    with pytest.deprecated_call():
        import ophyd_async.epics.signal  # noqa: F401



from ._aravis import AravisDetector
from ._aravis_controller import AravisController
from ._aravis_io import AravisDriverIO

__all__ = [
    "AravisDetector",
    "AravisController",
    "AravisDriverIO",
]



from collections.abc import Sequence
from os import path
from typing import Any
from unittest.mock import patch

import numpy as np
import numpy.typing as npt
import pytest
import yaml
from bluesky.run_engine import RunEngine

from ophyd_async.core import (
    Array1D,
    Device,
    SignalRW,
    StrictEnum,
    Table,
    all_at_once,
    get_signal_values,
    load_device,
    load_from_yaml,
    save_device,
    save_to_yaml,
    set_signal_values,
    walk_rw_signals,
)
from ophyd_async.epics.core import epics_signal_r, epics_signal_rw


class EnumTest(StrictEnum):
    VAL1 = "val1"
    VAL2 = "val2"


class DummyChildDevice(Device):
    def __init__(self) -> None:
        self.str_sig = epics_signal_rw(str, "StrSignal")
        super().__init__()


class DummyDeviceGroup(Device):
    def __init__(self, name: str):
        self.child1 = DummyChildDevice()
        self.child2 = DummyChildDevice()
        self.str_sig = epics_signal_rw(str, "ParentValue1")
        self.parent_sig2 = epics_signal_r(
            int, "ParentValue2"
        )  # Ensure only RW are found
        self.table_sig = epics_signal_rw(Table, "TableSignal")
        self.array_sig = epics_signal_rw(Array1D[np.uint32], "ArraySignal")
        self.enum_sig = epics_signal_rw(EnumTest, "EnumSignal")
        super().__init__(name)


class MyEnum(StrictEnum):
    one = "one"
    two = "two"
    three = "three"


class SomeTable(Table):
    some_float: Array1D[np.float64]
    some_int: Array1D[np.int32]
    some_enum: Sequence[MyEnum]


class DummyDeviceGroupAllTypes(Device):
    def __init__(self, name: str):
        self.pv_int: SignalRW = epics_signal_rw(int, "PV1")
        self.pv_float: SignalRW = epics_signal_rw(float, "PV2")
        self.pv_str: SignalRW = epics_signal_rw(str, "PV2")
        self.pv_enum_str: SignalRW = epics_signal_rw(MyEnum, "PV3")
        self.pv_enum: SignalRW = epics_signal_rw(MyEnum, "PV4")
        self.pv_array_int8 = epics_signal_rw(npt.NDArray[np.int8], "PV5")
        self.pv_array_uint8 = epics_signal_rw(npt.NDArray[np.uint8], "PV6")
        self.pv_array_int16 = epics_signal_rw(npt.NDArray[np.int16], "PV7")
        self.pv_array_uint16 = epics_signal_rw(npt.NDArray[np.uint16], "PV8")
        self.pv_array_int32 = epics_signal_rw(npt.NDArray[np.int32], "PV9")
        self.pv_array_uint32 = epics_signal_rw(npt.NDArray[np.uint32], "PV10")
        self.pv_array_int64 = epics_signal_rw(npt.NDArray[np.int64], "PV11")
        self.pv_array_uint64 = epics_signal_rw(npt.NDArray[np.uint64], "PV12")
        self.pv_array_float32 = epics_signal_rw(npt.NDArray[np.float32], "PV13")
        self.pv_array_float64 = epics_signal_rw(npt.NDArray[np.float64], "PV14")
        self.pv_array_npstr = epics_signal_rw(npt.NDArray[np.str_], "PV15")
        self.pv_array_str = epics_signal_rw(Sequence[str], "PV16")
        self.pv_protocol_device_abstraction = epics_signal_rw(Table, "pva://PV17")
        super().__init__(name)


@pytest.fixture
async def device() -> DummyDeviceGroup:
    device = DummyDeviceGroup("parent")
    await device.connect(mock=True)
    return device


@pytest.fixture
async def device_all_types() -> DummyDeviceGroupAllTypes:
    device = DummyDeviceGroupAllTypes("parent")
    await device.connect(mock=True)
    return device


# Dummy function to check different phases save properly
def sort_signal_by_phase(values: dict[str, Any]) -> list[dict[str, Any]]:
    phase_1 = {"child1.str_sig": values["child1.str_sig"]}
    phase_2 = {"child2.str_sig": values["child2.str_sig"]}
    phase_3 = {
        key: value
        for key, value in values.items()
        if key not in phase_1 and key not in phase_2
    }
    return [phase_1, phase_2, phase_3]


async def test_enum_yaml_formatting(tmp_path):
    enums = [EnumTest.VAL1, EnumTest.VAL2]
    save_to_yaml(enums, path.join(tmp_path, "test_file.yaml"))
    with open(path.join(tmp_path, "test_file.yaml")) as file:
        saved_enums = yaml.load(file, yaml.Loader)
    # check that save/load reduces from enum to str
    assert all(isinstance(value, str) for value in saved_enums)
    # check values of enums same
    assert saved_enums == enums


async def test_save_device_all_types(
    RE: RunEngine, device_all_types: DummyDeviceGroupAllTypes, tmp_path
):
    # Populate fake device with PV's...
    await device_all_types.pv_int.set(1)
    await device_all_types.pv_float.set(1.234)
    await device_all_types.pv_str.set("test_string")
    await device_all_types.pv_enum_str.set("two")
    await device_all_types.pv_enum.set(MyEnum.two)
    for pv, dtype in {
        device_all_types.pv_array_int8: np.int8,
        device_all_types.pv_array_uint8: np.uint8,
        device_all_types.pv_array_int16: np.int16,
        device_all_types.pv_array_uint16: np.uint16,
        device_all_types.pv_array_int32: np.int32,
        device_all_types.pv_array_uint32: np.uint32,
        device_all_types.pv_array_int64: np.int64,
        device_all_types.pv_array_uint64: np.uint64,
    }.items():
        await pv.set(
            np.array(
                [np.iinfo(dtype).min, np.iinfo(dtype).max, 0, 1, 2, 3, 4], dtype=dtype
            )
        )
    for pv, dtype in {
        device_all_types.pv_array_float32: np.float32,
        device_all_types.pv_array_float64: np.float64,
    }.items():
        finfo = np.finfo(dtype)
        data = np.array(
            [
                finfo.min,
                finfo.max,
                finfo.smallest_normal,
                finfo.smallest_subnormal,
                0,
                1.234,
                2.34e5,
                3.45e-6,
            ],
            dtype=dtype,
        )

        await pv.set(data)
    await device_all_types.pv_array_npstr.set(
        np.array(["one", "two", "three"], dtype=np.str_),
    )
    await device_all_types.pv_array_str.set(
        ["one", "two", "three"],
    )
    await device_all_types.pv_protocol_device_abstraction.set(
        SomeTable(
            some_float=np.arange(3, dtype=np.float64),
            some_int=np.arange(3),
            some_enum=[MyEnum.one, MyEnum.two, MyEnum.three],
        )
    )

    # Create save plan from utility functions
    def save_my_device():
        signalRWs = walk_rw_signals(device_all_types)
        values = yield from get_signal_values(signalRWs)

        save_to_yaml([values], path.join(tmp_path, "test_file.yaml"))

    RE(save_my_device())

    actual_file_path = path.join(tmp_path, "test_file.yaml")
    with open(actual_file_path) as actual_file:
        with open("tests/test_data/test_yaml_save.yml") as expected_file:
            assert yaml.safe_load(actual_file) == yaml.safe_load(expected_file)


async def test_save_device(RE: RunEngine, device: DummyDeviceGroup, tmp_path):
    # Populate fake device with PV's...
    await device.child1.str_sig.set("test_string")
    # Test tables PVs
    table_pv = {"VAL1": np.array([1, 1, 1, 1, 1]), "VAL2": np.array([1, 1, 1, 1, 1])}
    array_pv = np.array([2, 2, 2, 2, 2])
    await device.array_sig.set(array_pv)
    await device.table_sig.set(table_pv)
    await device.enum_sig.set(EnumTest.VAL2)

    # Create save plan from utility functions
    def save_my_device():
        signalRWs = walk_rw_signals(device)

        assert list(signalRWs.keys()) == [
            "child1.str_sig",
            "child2.str_sig",
            "str_sig",
            "table_sig",
            "array_sig",
            "enum_sig",
        ]
        assert all(isinstance(signal, SignalRW) for signal in list(signalRWs.values()))

        values = yield from get_signal_values(signalRWs, ignore=["str_sig"])
        assert np.array_equal(values["array_sig"], array_pv)
        assert values["enum_sig"] == "val2"
        assert values["table_sig"] == Table(**table_pv)
        assert values["str_sig"] is None
        assert values["child1.str_sig"] == "test_string"
        assert values["child2.str_sig"] == ""

        save_to_yaml([values], path.join(tmp_path, "test_file.yaml"))

    RE(save_my_device())

    with open(path.join(tmp_path, "test_file.yaml")) as file:
        yaml_content = yaml.load(file, yaml.Loader)[0]
        assert yaml_content["child1.str_sig"] == "test_string"
        assert yaml_content["child2.str_sig"] == ""
        assert np.array_equal(yaml_content["table_sig"]["VAL1"], table_pv["VAL1"])
        assert np.array_equal(yaml_content["table_sig"]["VAL2"], table_pv["VAL2"])
        assert np.array_equal(yaml_content["array_sig"], array_pv)
        assert yaml_content["enum_sig"] == "val2"
        assert yaml_content["str_sig"] is None


async def test_yaml_formatting(RE: RunEngine, device: DummyDeviceGroup, tmp_path):
    file_path = path.join(tmp_path, "test_file.yaml")
    await device.child1.str_sig.set("test_string")
    table = {"VAL1": np.array([1, 2, 3, 4, 5]), "VAL2": np.array([6, 7, 8, 9, 10])}
    await device.array_sig.set(np.array([11, 12, 13, 14, 15]))
    await device.table_sig.set(table)
    await device.enum_sig.set(EnumTest.VAL2)
    RE(save_device(device, file_path, sorter=sort_signal_by_phase))

    with open(file_path) as file:
        expected = """\
- child1.str_sig: test_string
- child2.str_sig: ''
- array_sig: [11, 12, 13, 14, 15]
  enum_sig: val2
  str_sig: ''
  table_sig:
    VAL1: [1, 2, 3, 4, 5]
    VAL2: [6, 7, 8, 9, 10]
"""
        # assert False, file.read()
        assert file.read() == expected


async def test_load_from_yaml(RE: RunEngine, device: DummyDeviceGroup, tmp_path):
    file_path = path.join(tmp_path, "test_file.yaml")

    array = np.array([1, 1, 1, 1, 1])
    table = {"VAL1": np.array([1, 2, 3, 4, 5]), "VAL2": np.array([6, 7, 8, 9, 10])}
    await device.child1.str_sig.set("initial_string")
    await device.array_sig.set(array)
    await device.str_sig.set(None)
    await device.enum_sig.set(EnumTest.VAL2)
    await device.table_sig.set(table)
    RE(save_device(device, file_path, sorter=sort_signal_by_phase))

    values = load_from_yaml(file_path)
    assert values[0]["child1.str_sig"] == "initial_string"
    assert values[1]["child2.str_sig"] == ""
    assert values[2]["str_sig"] == ""
    assert values[2]["enum_sig"] == "val2"
    assert np.array_equal(values[2]["array_sig"], array)
    assert np.array_equal(values[2]["table_sig"]["VAL1"], table["VAL1"])
    assert np.array_equal(values[2]["table_sig"]["VAL2"], table["VAL2"])


async def test_set_signal_values_restores_value(
    RE: RunEngine, device: DummyDeviceGroup, tmp_path
):
    file_path = path.join(tmp_path, "test_file.yaml")

    await device.str_sig.set("initial_string")
    await device.array_sig.set(np.array([1, 1, 1, 1, 1]))
    RE(save_device(device, file_path, sorter=sort_signal_by_phase))

    await device.str_sig.set("changed_string")
    await device.array_sig.set(np.array([2, 2, 2, 2, 2]))
    string_value = await device.str_sig.get_value()
    array_value = await device.array_sig.get_value()
    assert string_value == "changed_string"
    assert np.array_equal(array_value, np.array([2, 2, 2, 2, 2]))

    values = load_from_yaml(file_path)
    signals_to_set = walk_rw_signals(device)

    RE(set_signal_values(signals_to_set, values))

    string_value = await device.str_sig.get_value()
    array_value = await device.array_sig.get_value()
    assert string_value == "initial_string"
    assert np.array_equal(array_value, np.array([1, 1, 1, 1, 1]))


@patch("ophyd_async.core._device_save_loader.load_from_yaml")
@patch("ophyd_async.core._device_save_loader.walk_rw_signals")
@patch("ophyd_async.core._device_save_loader.set_signal_values")
async def test_load_device(
    mock_set_signal_values,
    mock_walk_rw_signals,
    mock_load_from_yaml,
    device: DummyDeviceGroup,
):
    RE = RunEngine()
    RE(load_device(device, "path"))
    mock_load_from_yaml.assert_called_once()
    mock_walk_rw_signals.assert_called_once()
    mock_set_signal_values.assert_called_once()


async def test_set_signal_values_skips_ignored_values(device: DummyDeviceGroup):
    RE = RunEngine()
    array = np.array([1, 1, 1, 1, 1])

    await device.child1.str_sig.set("initial_string")
    await device.array_sig.set(array)
    await device.str_sig.set(None)

    signals_of_device = walk_rw_signals(device)
    values_to_set = [{"child1.str_sig": None, "array_sig": np.array([2, 3, 4])}]

    RE(set_signal_values(signals_of_device, values_to_set))

    assert np.all(await device.array_sig.get_value() == np.array([2, 3, 4]))
    assert await device.child1.str_sig.get_value() == "initial_string"


def test_all_at_once_sorter():
    assert all_at_once({"child1.str_sig": 0}) == [{"child1.str_sig": 0}]



import asyncio
from collections.abc import AsyncGenerator, AsyncIterator
from pathlib import Path
from xml.etree import ElementTree as ET

from bluesky.protocols import Hints, StreamAsset
from event_model import DataKey

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    DatasetDescriber,
    DetectorWriter,
    HDFDataset,
    HDFFile,
    NameProvider,
    PathProvider,
    observe_value,
    set_and_wait_for_value,
    wait_for_value,
)

from ._core_io import NDArrayBaseIO, NDFileHDFIO
from ._utils import (
    FileWriteMode,
    convert_param_dtype_to_np,
    convert_pv_dtype_to_np,
)


class ADHDFWriter(DetectorWriter):
    def __init__(
        self,
        hdf: NDFileHDFIO,
        path_provider: PathProvider,
        name_provider: NameProvider,
        dataset_describer: DatasetDescriber,
        *plugins: NDArrayBaseIO,
    ) -> None:
        self.hdf = hdf
        self._path_provider = path_provider
        self._name_provider = name_provider
        self._dataset_describer = dataset_describer

        self._plugins = plugins
        self._capture_status: AsyncStatus | None = None
        self._datasets: list[HDFDataset] = []
        self._file: HDFFile | None = None
        self._multiplier = 1

    async def open(self, multiplier: int = 1) -> dict[str, DataKey]:
        self._file = None
        info = self._path_provider(device_name=self._name_provider())

        # Set the directory creation depth first, since dir creation callback happens
        # when directory path PV is processed.
        await self.hdf.create_directory.set(info.create_dir_depth)

        # Make sure we are using chunk auto-sizing
        await asyncio.gather(self.hdf.chunk_size_auto.set(True))

        await asyncio.gather(
            self.hdf.num_extra_dims.set(0),
            self.hdf.lazy_open.set(True),
            self.hdf.swmr_mode.set(True),
            # See https://github.com/bluesky/ophyd-async/issues/122
            self.hdf.file_path.set(str(info.directory_path)),
            self.hdf.file_name.set(info.filename),
            self.hdf.file_template.set("%s/%s.h5"),
            self.hdf.file_write_mode.set(FileWriteMode.stream),
            # Never use custom xml layout file but use the one defined
            # in the source code file NDFileHDF5LayoutXML.cpp
            self.hdf.xml_file_name.set(""),
        )

        assert (
            await self.hdf.file_path_exists.get_value()
        ), f"File path {info.directory_path} for hdf plugin does not exist"

        # Overwrite num_capture to go forever
        await self.hdf.num_capture.set(0)
        # Wait for it to start, stashing the status that tells us when it finishes
        self._capture_status = await set_and_wait_for_value(self.hdf.capture, True)
        name = self._name_provider()
        detector_shape = await self._dataset_describer.shape()
        np_dtype = await self._dataset_describer.np_datatype()
        self._multiplier = multiplier
        outer_shape = (multiplier,) if multiplier > 1 else ()

        # Determine number of frames that will be saved per HDF chunk
        frames_per_chunk = await self.hdf.num_frames_chunks.get_value()

        # Add the main data
        self._datasets = [
            HDFDataset(
                data_key=name,
                dataset="/entry/data/data",
                shape=detector_shape,
                dtype_numpy=np_dtype,
                multiplier=multiplier,
                chunk_shape=(frames_per_chunk, *detector_shape),
            )
        ]
        # And all the scalar datasets
        for plugin in self._plugins:
            maybe_xml = await plugin.nd_attributes_file.get_value()
            # This is the check that ADCore does to see if it is an XML string
            # rather than a filename to parse
            if "<Attributes>" in maybe_xml:
                root = ET.fromstring(maybe_xml)
                for child in root:
                    datakey = child.attrib["name"]
                    if child.attrib.get("type", "EPICS_PV") == "EPICS_PV":
                        np_datatype = convert_pv_dtype_to_np(
                            child.attrib.get("dbrtype", "DBR_NATIVE")
                        )
                    else:
                        np_datatype = convert_param_dtype_to_np(
                            child.attrib.get("datatype", "INT")
                        )
                    self._datasets.append(
                        HDFDataset(
                            datakey,
                            f"/entry/instrument/NDAttributes/{datakey}",
                            (),
                            np_datatype,
                            multiplier,
                            # NDAttributes appear to always be configured with
                            # this chunk size
                            chunk_shape=(16384,),
                        )
                    )

        describe = {
            ds.data_key: DataKey(
                source=self.hdf.full_file_name.source,
                shape=list(outer_shape + tuple(ds.shape)),
                dtype="array" if ds.shape else "number",
                dtype_numpy=ds.dtype_numpy,
                external="STREAM:",
            )
            for ds in self._datasets
        }
        return describe

    async def observe_indices_written(
        self, timeout=DEFAULT_TIMEOUT
    ) -> AsyncGenerator[int, None]:
        """Wait until a specific index is ready to be collected"""
        async for num_captured in observe_value(self.hdf.num_captured, timeout):
            yield num_captured // self._multiplier

    async def get_indices_written(self) -> int:
        num_captured = await self.hdf.num_captured.get_value()
        return num_captured // self._multiplier

    async def collect_stream_docs(
        self, indices_written: int
    ) -> AsyncIterator[StreamAsset]:
        # TODO: fail if we get dropped frames
        await self.hdf.flush_now.set(True)
        if indices_written:
            if not self._file:
                path = Path(await self.hdf.full_file_name.get_value())
                self._file = HDFFile(
                    # See https://github.com/bluesky/ophyd-async/issues/122
                    path,
                    self._datasets,
                )
                # stream resource says "here is a dataset",
                # stream datum says "here are N frames in that stream resource",
                # you get one stream resource and many stream datums per scan

                for doc in self._file.stream_resources():
                    yield "stream_resource", doc
            for doc in self._file.stream_data(indices_written):
                yield "stream_datum", doc

    async def close(self):
        # Already done a caput callback in _capture_status, so can't do one here
        await self.hdf.capture.set(False, wait=False)
        await wait_for_value(self.hdf.capture, False, DEFAULT_TIMEOUT)
        if self._capture_status:
            # We kicked off an open, so wait for it to return
            await self._capture_status

    @property
    def hints(self) -> Hints:
        return {"fields": [self._name_provider()]}



import asyncio
from collections.abc import Callable
from functools import cached_property
from unittest.mock import AsyncMock, Mock

from bluesky.protocols import Descriptor, Reading

from ._signal_backend import SignalBackend, SignalDatatypeT
from ._soft_signal_backend import SoftSignalBackend
from ._utils import Callback


class MockSignalBackend(SignalBackend[SignalDatatypeT]):
    """Signal backend for testing, created by ``Device.connect(mock=True)``."""

    def __init__(
        self,
        initial_backend: SignalBackend[SignalDatatypeT],
        mock: Mock,
    ) -> None:
        if isinstance(initial_backend, MockSignalBackend):
            raise ValueError("Cannot make a MockSignalBackend for a MockSignalBackend")

        self.initial_backend = initial_backend

        if isinstance(self.initial_backend, SoftSignalBackend):
            # Backend is already a SoftSignalBackend, so use it
            self.soft_backend = self.initial_backend
        else:
            # Backend is not a SoftSignalBackend, so create one to mimic it
            self.soft_backend = SoftSignalBackend(
                datatype=self.initial_backend.datatype
            )

        # use existing Mock if provided
        self.mock = mock
        self.put_mock = AsyncMock(name="put", spec=Callable)
        self.mock.attach_mock(self.put_mock, "put")

        super().__init__(datatype=self.initial_backend.datatype)

    def set_value(self, value: SignalDatatypeT):
        self.soft_backend.set_value(value)

    def source(self, name: str, read: bool) -> str:
        return f"mock+{self.initial_backend.source(name, read)}"

    async def connect(self, timeout: float) -> None:
        pass

    @cached_property
    def put_proceeds(self) -> asyncio.Event:
        put_proceeds = asyncio.Event()
        put_proceeds.set()
        return put_proceeds

    async def put(self, value: SignalDatatypeT | None, wait: bool):
        await self.put_mock(value, wait=wait)
        await self.soft_backend.put(value, wait=wait)
        if wait:
            await self.put_proceeds.wait()

    async def get_reading(self) -> Reading:
        return await self.soft_backend.get_reading()

    async def get_value(self) -> SignalDatatypeT:
        return await self.soft_backend.get_value()

    async def get_setpoint(self) -> SignalDatatypeT:
        return await self.soft_backend.get_setpoint()

    async def get_datakey(self, source: str) -> Descriptor:
        return await self.soft_backend.get_datakey(source)

    def set_callback(self, callback: Callback[Reading[SignalDatatypeT]] | None) -> None:
        self.soft_backend.set_callback(callback)



import os
from unittest.mock import ANY

import bluesky.plan_stubs as bps
import pytest
from bluesky import RunEngine

from ophyd_async.core import (
    StandardFlyer,
    StaticFilenameProvider,
    StaticPathProvider,
    assert_emitted,
    callback_on_mock_put,
    set_mock_value,
)
from ophyd_async.fastcs.panda import (
    DatasetTable,
    HDFPanda,
    PandaHdf5DatasetType,
    StaticSeqTableTriggerLogic,
)
from ophyd_async.plan_stubs import (
    prepare_static_seq_table_flyer_and_detectors_with_same_trigger,
)


@pytest.fixture
async def mock_hdf_panda(tmp_path):
    fp = StaticFilenameProvider("test-panda")
    dp = StaticPathProvider(fp, tmp_path)

    mock_hdf_panda = HDFPanda("HDFPANDA:", path_provider=dp, name="panda")
    await mock_hdf_panda.connect(mock=True)

    def link_function(value, **kwargs):
        set_mock_value(mock_hdf_panda.pcap.active, value)

    # Mimic directory exists check that happens normally in the PandA IOC
    def check_dir_exits(value, **kwargs):
        if os.path.exists(value):
            set_mock_value(mock_hdf_panda.data.directory_exists, True)

    callback_on_mock_put(mock_hdf_panda.pcap.arm, link_function)
    callback_on_mock_put(mock_hdf_panda.data.hdf_directory, check_dir_exits)

    set_mock_value(
        mock_hdf_panda.data.datasets,
        DatasetTable(
            name=["x", "y"],
            hdf5_type=[PandaHdf5DatasetType.UINT_32, PandaHdf5DatasetType.FLOAT_64],
        ),
    )

    yield mock_hdf_panda


async def test_hdf_panda_passes_blocks_to_controller(mock_hdf_panda: HDFPanda):
    assert hasattr(mock_hdf_panda.controller, "pcap")
    assert mock_hdf_panda.controller.pcap is mock_hdf_panda.pcap


async def test_hdf_panda_hardware_triggered_flyable(
    RE: RunEngine,
    mock_hdf_panda,
    tmp_path,
):
    docs = {}

    def append_and_print(name, doc):
        if name not in docs:
            docs[name] = []
        docs[name] += [doc]

    RE.subscribe(append_and_print)

    shutter_time = 0.004
    exposure = 1

    trigger_logic = StaticSeqTableTriggerLogic(mock_hdf_panda.seq[1])
    flyer = StandardFlyer(trigger_logic, name="flyer")

    def flying_plan():
        yield from bps.stage_all(mock_hdf_panda, flyer)

        yield from prepare_static_seq_table_flyer_and_detectors_with_same_trigger(
            flyer,
            [mock_hdf_panda],
            number_of_frames=1,
            exposure=exposure,
            shutter_time=shutter_time,
        )

        yield from bps.open_run()
        yield from bps.declare_stream(mock_hdf_panda, name="main_stream", collect=True)

        set_mock_value(flyer.trigger_logic.seq.active, 1)

        yield from bps.kickoff(flyer, wait=True)
        yield from bps.kickoff(mock_hdf_panda, wait=True)

        yield from bps.complete(flyer, wait=False, group="complete")
        yield from bps.complete(mock_hdf_panda, wait=False, group="complete")

        # Manually incremenet the index as if a frame was taken
        set_mock_value(mock_hdf_panda.data.num_captured, 1)
        set_mock_value(flyer.trigger_logic.seq.active, 0)

        done = False
        while not done:
            try:
                yield from bps.wait(group="complete", timeout=0.5)
            except TimeoutError:
                pass
            else:
                done = True
            yield from bps.collect(
                mock_hdf_panda,
                return_payload=False,
                name="main_stream",
            )
        yield from bps.wait(group="complete")
        yield from bps.close_run()
        # Verify that _completable_frames is reset to 0 after the final complete.
        assert mock_hdf_panda._completable_frames == 0
        yield from bps.unstage_all(flyer, mock_hdf_panda)
        yield from bps.wait_for([lambda: mock_hdf_panda.controller.disarm()])

    # fly scan
    RE(flying_plan())

    assert_emitted(
        docs, start=1, descriptor=1, stream_resource=2, stream_datum=2, stop=1
    )

    # test descriptor
    data_key_names: dict[str, str] = docs["descriptor"][0]["object_keys"]["panda"]
    assert data_key_names == ["x", "y"]
    for data_key_name in data_key_names:
        assert (
            docs["descriptor"][0]["data_keys"][data_key_name]["source"] == "mock+pva://"
        )

    # test stream resources
    for dataset_name, stream_resource, data_key_name in zip(
        ("x", "y"), docs["stream_resource"], data_key_names, strict=False
    ):
        assert stream_resource == {
            "run_start": docs["start"][0]["uid"],
            "uid": ANY,
            "data_key": data_key_name,
            "mimetype": "application/x-hdf5",
            "uri": "file://localhost/" + str(tmp_path / "test-panda.h5").lstrip("/"),
            "parameters": {
                "dataset": f"/{dataset_name}",
                "swmr": False,
                "multiplier": 1,
                "chunk_shape": (1024,),
            },
        }
        assert "test-panda.h5" in stream_resource["uri"]

    # test stream datum
    for stream_datum in docs["stream_datum"]:
        assert stream_datum["descriptor"] == docs["descriptor"][0]["uid"]
        assert stream_datum["seq_nums"] == {
            "start": 1,
            "stop": 2,
        }
        assert stream_datum["indices"] == {
            "start": 0,
            "stop": 1,
        }
        assert stream_datum["stream_resource"] in [
            sd["uid"].split("/")[0] for sd in docs["stream_datum"]
        ]



import os
from collections.abc import Callable

import pytest
from bluesky.run_engine import RunEngine

from ophyd_async.core import (
    DeviceCollector,
    StandardDetector,
    callback_on_mock_put,
    set_mock_value,
)


@pytest.fixture
def ad_standard_det_factory(
    RE: RunEngine,
    static_path_provider,
) -> Callable[[StandardDetector, int], StandardDetector]:
    def generate_ad_standard_det(
        ad_standard_detector_class, number=1
    ) -> StandardDetector:
        # Dynamically generate a name based on the class of detector
        detector_name = ad_standard_detector_class.__name__
        if detector_name.endswith("Detector"):
            detector_name = detector_name[: -len("Detector")]

        with DeviceCollector(mock=True):
            test_adstandard_det = ad_standard_detector_class(
                f"{detector_name.upper()}{number}:",
                static_path_provider,
                name=f"test_ad{detector_name.lower()}{number}",
            )

        def on_set_file_path_callback(value, **kwargs):
            if os.path.exists(value):
                set_mock_value(test_adstandard_det.hdf.file_path_exists, True)
                set_mock_value(
                    test_adstandard_det.hdf.full_file_name,
                    f"{value}/{static_path_provider._filename_provider(device_name=test_adstandard_det.name)}.h5",
                )

        callback_on_mock_put(
            test_adstandard_det.hdf.file_path, on_set_file_path_callback
        )

        # Set some sensible defaults to mimic a real detector setup
        set_mock_value(test_adstandard_det.drv.acquire_time, (number - 0.2))
        set_mock_value(test_adstandard_det.drv.acquire_period, float(number))
        set_mock_value(test_adstandard_det.hdf.capture, True)

        # Set number of frames per chunk and frame dimensions to something reasonable
        set_mock_value(test_adstandard_det.hdf.num_frames_chunks, 1)
        set_mock_value(test_adstandard_det.drv.array_size_x, (9 + number))
        set_mock_value(test_adstandard_det.drv.array_size_y, (9 + number))

        return test_adstandard_det

    return generate_ad_standard_det



from ._eiger import EigerDetector, EigerTriggerInfo
from ._eiger_controller import EigerController
from ._eiger_io import EigerDriverIO

__all__ = ["EigerDetector", "EigerController", "EigerDriverIO", "EigerTriggerInfo"]



from dataclasses import dataclass

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    SignalDatatypeT,
    SignalR,
    SignalRW,
    StrictEnum,
    wait_for_value,
)


class ADBaseDataType(StrictEnum):
    Int8 = "Int8"
    UInt8 = "UInt8"
    Int16 = "Int16"
    UInt16 = "UInt16"
    Int32 = "Int32"
    UInt32 = "UInt32"
    Int64 = "Int64"
    UInt64 = "UInt64"
    Float32 = "Float32"
    Float64 = "Float64"


def convert_ad_dtype_to_np(ad_dtype: ADBaseDataType) -> str:
    ad_dtype_to_np_dtype = {
        ADBaseDataType.Int8: "|i1",
        ADBaseDataType.UInt8: "|u1",
        ADBaseDataType.Int16: "<i2",
        ADBaseDataType.UInt16: "<u2",
        ADBaseDataType.Int32: "<i4",
        ADBaseDataType.UInt32: "<u4",
        ADBaseDataType.Int64: "<i8",
        ADBaseDataType.UInt64: "<u8",
        ADBaseDataType.Float32: "<f4",
        ADBaseDataType.Float64: "<f8",
    }
    return ad_dtype_to_np_dtype[ad_dtype]


def convert_pv_dtype_to_np(datatype: str) -> str:
    _pvattribute_to_ad_datatype = {
        "DBR_SHORT": ADBaseDataType.Int16,
        "DBR_ENUM": ADBaseDataType.Int16,
        "DBR_INT": ADBaseDataType.Int32,
        "DBR_LONG": ADBaseDataType.Int32,
        "DBR_FLOAT": ADBaseDataType.Float32,
        "DBR_DOUBLE": ADBaseDataType.Float64,
    }
    if datatype in ["DBR_STRING", "DBR_CHAR"]:
        np_datatype = "s40"
    elif datatype == "DBR_NATIVE":
        raise ValueError("Don't support DBR_NATIVE yet")
    else:
        try:
            np_datatype = convert_ad_dtype_to_np(_pvattribute_to_ad_datatype[datatype])
        except KeyError as e:
            raise ValueError(f"Invalid dbr type {datatype}") from e
    return np_datatype


def convert_param_dtype_to_np(datatype: str) -> str:
    _paramattribute_to_ad_datatype = {
        "INT": ADBaseDataType.Int32,
        "INT64": ADBaseDataType.Int64,
        "DOUBLE": ADBaseDataType.Float64,
    }
    if datatype in ["STRING"]:
        np_datatype = "s40"
    else:
        try:
            np_datatype = convert_ad_dtype_to_np(
                _paramattribute_to_ad_datatype[datatype]
            )
        except KeyError as e:
            raise ValueError(f"Invalid datatype {datatype}") from e
    return np_datatype


class FileWriteMode(StrictEnum):
    single = "Single"
    capture = "Capture"
    stream = "Stream"


class ImageMode(StrictEnum):
    single = "Single"
    multiple = "Multiple"
    continuous = "Continuous"


class NDAttributeDataType(StrictEnum):
    INT = "INT"
    DOUBLE = "DOUBLE"
    STRING = "STRING"


class NDAttributePvDbrType(StrictEnum):
    DBR_SHORT = "DBR_SHORT"
    DBR_ENUM = "DBR_ENUM"
    DBR_INT = "DBR_INT"
    DBR_LONG = "DBR_LONG"
    DBR_FLOAT = "DBR_FLOAT"
    DBR_DOUBLE = "DBR_DOUBLE"
    DBR_STRING = "DBR_STRING"
    DBR_CHAR = "DBR_CHAR"


@dataclass
class NDAttributePv:
    name: str  # name of attribute stamped on array, also scientifically useful name
    # when appended to device.name
    signal: SignalR  # caget the pv given by signal.source and attach to each frame
    dbrtype: NDAttributePvDbrType
    description: str = ""  # A description that appears in the HDF file as an attribute


@dataclass
class NDAttributeParam:
    name: str  # name of attribute stamped on array, also scientifically useful name
    # when appended to device.name
    param: str  # The parameter string as seen in the INP link of the record
    datatype: NDAttributeDataType  # The datatype of the parameter
    addr: int = 0  # The address as seen in the INP link of the record
    description: str = ""  # A description that appears in the HDF file as an attribute


async def stop_busy_record(
    signal: SignalRW[SignalDatatypeT],
    value: SignalDatatypeT,
    timeout: float = DEFAULT_TIMEOUT,
    status_timeout: float | None = None,
) -> None:
    await signal.set(value, wait=False, timeout=status_timeout)
    await wait_for_value(signal, value, timeout=timeout)



from unittest.mock import ANY, patch

from pytest import fixture, raises

from ophyd_async.core import (
    DeviceCollector,
    TriggerInfo,
    callback_on_mock_put,
    get_mock_put,
    set_mock_value,
)
from ophyd_async.epics.eiger import EigerController, EigerDriverIO

DriverAndController = tuple[EigerDriverIO, EigerController]


@fixture
def eiger_driver_and_controller_no_arm(RE) -> DriverAndController:
    with DeviceCollector(mock=True):
        driver = EigerDriverIO("")
        controller = EigerController(driver)

    return driver, controller


@fixture
def eiger_driver_and_controller(
    eiger_driver_and_controller_no_arm: DriverAndController,
) -> DriverAndController:
    driver, controller = eiger_driver_and_controller_no_arm

    def become_ready_on_arm(*args, **kwargs):
        if args[0] == 1:
            set_mock_value(driver.state, "ready")

    callback_on_mock_put(driver.arm, become_ready_on_arm)

    return driver, controller


async def test_when_arm_with_exposure_then_time_and_period_set(
    eiger_driver_and_controller: DriverAndController,
):
    driver, controller = eiger_driver_and_controller
    test_exposure = 0.002
    await controller.prepare(TriggerInfo(number_of_triggers=10, livetime=test_exposure))
    await controller.arm()
    await controller.wait_for_idle()
    assert (await driver.acquire_period.get_value()) == test_exposure
    assert (await driver.acquire_time.get_value()) == test_exposure


async def test_when_arm_with_no_exposure_then_arm_set_correctly(
    eiger_driver_and_controller: DriverAndController,
):
    driver, controller = eiger_driver_and_controller
    await controller.prepare(TriggerInfo(number_of_triggers=10))
    await controller.arm()
    await controller.wait_for_idle()
    get_mock_put(driver.arm).assert_called_once_with(1, wait=ANY)


async def test_when_arm_with_number_of_images_then_number_of_images_set_correctly(
    eiger_driver_and_controller: DriverAndController,
):
    driver, controller = eiger_driver_and_controller
    test_number_of_images = 40
    await controller.prepare(TriggerInfo(number_of_triggers=test_number_of_images))
    await controller.arm()
    await controller.wait_for_idle()
    get_mock_put(driver.num_images).assert_called_once_with(
        test_number_of_images, wait=ANY
    )


@patch("ophyd_async.epics.eiger._eiger_controller.DEFAULT_TIMEOUT", 0.1)
async def test_given_detector_fails_to_go_ready_when_arm_called_then_fails(
    eiger_driver_and_controller_no_arm: DriverAndController,
):
    _, controller = eiger_driver_and_controller_no_arm
    with raises(TimeoutError):
        await controller.prepare(TriggerInfo(number_of_triggers=10))
        await controller.arm()
        await controller.wait_for_idle()


async def test_when_disarm_called_on_controller_then_disarm_called_on_driver(
    eiger_driver_and_controller: DriverAndController,
):
    driver, controller = eiger_driver_and_controller
    await controller.disarm()
    get_mock_put(driver.disarm).assert_called_once_with(1, wait=ANY)


async def test_when_get_deadtime_called_then_returns_expected_deadtime(
    eiger_driver_and_controller: DriverAndController,
):
    _, controller = eiger_driver_and_controller
    assert controller.get_deadtime(0) == 0.0001


async def test_given_energy_within_tolerance_when_photon_energy_set_then_pv_unchanged(
    eiger_driver_and_controller: DriverAndController,
):
    driver, controller = eiger_driver_and_controller
    initial_energy = 10
    set_mock_value(driver.photon_energy, initial_energy)
    await controller.set_energy(10.002)
    get_mock_put(driver.photon_energy).assert_not_called()
    assert (await driver.photon_energy.get_value()) == initial_energy


async def test_given_energy_outside_tolerance_when_photon_energy_set_then_pv_changed(
    eiger_driver_and_controller: DriverAndController,
):
    driver, controller = eiger_driver_and_controller
    initial_energy = 10
    new_energy = 15
    set_mock_value(driver.photon_energy, initial_energy)
    await controller.set_energy(new_energy)
    get_mock_put(driver.photon_energy).assert_called_once()
    assert (await driver.photon_energy.get_value()) == new_energy



import asyncio
import os
from pathlib import Path

import h5py
import pytest
from bluesky.run_engine import RunEngine

from ophyd_async.core import (
    DetectorTrigger,
    Device,
    DeviceCollector,
    StaticPathProvider,
)
from ophyd_async.epics.core import epics_signal_rw
from ophyd_async.epics.eiger import EigerDetector, EigerTriggerInfo

SAVE_PATH = "/tmp"


class SetupDevice(Device):
    """Holds PVs that we would either expect to be initially set and
    never change or to be externally set in prod."""

    def __init__(self, eiger_prefix: str, odin_prefix: str) -> None:
        self.trigger = epics_signal_rw(int, f"{eiger_prefix}Trigger")
        self.header_detail = epics_signal_rw(str, f"{eiger_prefix}HeaderDetail")
        self.compression = epics_signal_rw(str, f"{odin_prefix}DatasetDataCompression")
        self.frames_per_block = epics_signal_rw(
            int, f"{odin_prefix}ProcessFramesPerBlock"
        )
        self.blocks_per_file = epics_signal_rw(
            int, f"{odin_prefix}ProcessBlocksPerFile"
        )
        super().__init__("")


@pytest.fixture
def ioc_prefixes():
    return os.environ["eiger_ioc"] + ":", os.environ["odin_ioc"] + ":"


@pytest.fixture
def RE():
    return RunEngine()


@pytest.fixture
async def setup_device(RE, ioc_prefixes):
    async with DeviceCollector():
        device = SetupDevice(ioc_prefixes[0], ioc_prefixes[1] + "FP:")
    await asyncio.gather(
        device.header_detail.set("all"),
        device.compression.set("BSLZ4"),
        device.frames_per_block.set(1000),
        device.blocks_per_file.set(1),
    )

    return device


@pytest.fixture
async def test_eiger(RE, ioc_prefixes) -> EigerDetector:
    provider = StaticPathProvider(lambda: "test_eiger", Path(SAVE_PATH))
    async with DeviceCollector():
        test_eiger = EigerDetector("", provider, ioc_prefixes[0], ioc_prefixes[1])

    return test_eiger


async def test_trigger_saves_file(test_eiger: EigerDetector, setup_device: SetupDevice):
    single_shot = EigerTriggerInfo(
        frame_timeout=None,
        number_of_triggers=1,
        trigger=DetectorTrigger.internal,
        deadtime=None,
        livetime=None,
        energy_ev=10000,
    )

    await test_eiger.stage()
    await test_eiger.prepare(single_shot)
    # Need to work out what the hold up is in prepare so we cant do this straight away.
    # File path propogation?
    await asyncio.sleep(0.5)
    await setup_device.trigger.set(1)
    await asyncio.sleep(0.5)  # Need to work out when it's actually finished writing
    await test_eiger.unstage()

    with h5py.File(SAVE_PATH + "/test_eiger_000001.h5") as f:
        assert "data" in f.keys()
        assert len(f["data"]) == 1



import pytest

from ophyd_async.core import (
    DetectorTrigger,
    PathProvider,
    TriggerInfo,
    set_mock_value,
)
from ophyd_async.epics import advimba
from ophyd_async.epics.advimba import (
    VimbaExposeOutMode,
    VimbaOnOff,
    VimbaTriggerSource,
)


@pytest.fixture
def test_advimba(ad_standard_det_factory) -> advimba.VimbaDetector:
    return ad_standard_det_factory(advimba.VimbaDetector)


async def test_get_deadtime(
    test_advimba: advimba.VimbaDetector,
):
    # Currently Vimba controller doesn't support getting deadtime.
    assert test_advimba._controller.get_deadtime(0) == 0.001


async def test_arming_trig_modes(test_advimba: advimba.VimbaDetector):
    set_mock_value(test_advimba.drv.trigger_source, VimbaTriggerSource.freerun)
    set_mock_value(test_advimba.drv.trigger_mode, VimbaOnOff.off)
    set_mock_value(test_advimba.drv.exposure_mode, VimbaExposeOutMode.timed)

    async def setup_trigger_mode(trig_mode: DetectorTrigger):
        await test_advimba.controller.prepare(
            TriggerInfo(number_of_triggers=1, trigger=trig_mode)
        )
        await test_advimba.controller.arm()
        await test_advimba.controller.wait_for_idle()
        # Prevent timeouts
        set_mock_value(test_advimba.drv.acquire, True)

    # Default TriggerSource
    assert (await test_advimba.drv.trigger_source.get_value()) == "Freerun"
    assert (await test_advimba.drv.trigger_mode.get_value()) == "Off"
    assert (await test_advimba.drv.exposure_mode.get_value()) == "Timed"

    await setup_trigger_mode(DetectorTrigger.edge_trigger)
    assert (await test_advimba.drv.trigger_source.get_value()) == "Line1"
    assert (await test_advimba.drv.trigger_mode.get_value()) == "On"
    assert (await test_advimba.drv.exposure_mode.get_value()) == "Timed"

    await setup_trigger_mode(DetectorTrigger.constant_gate)
    assert (await test_advimba.drv.trigger_source.get_value()) == "Line1"
    assert (await test_advimba.drv.trigger_mode.get_value()) == "On"
    assert (await test_advimba.drv.exposure_mode.get_value()) == "TriggerWidth"

    await setup_trigger_mode(DetectorTrigger.internal)
    assert (await test_advimba.drv.trigger_source.get_value()) == "Freerun"
    assert (await test_advimba.drv.trigger_mode.get_value()) == "Off"
    assert (await test_advimba.drv.exposure_mode.get_value()) == "Timed"

    await setup_trigger_mode(DetectorTrigger.variable_gate)
    assert (await test_advimba.drv.trigger_source.get_value()) == "Line1"
    assert (await test_advimba.drv.trigger_mode.get_value()) == "On"
    assert (await test_advimba.drv.exposure_mode.get_value()) == "TriggerWidth"


async def test_hints_from_hdf_writer(test_advimba: advimba.VimbaDetector):
    assert test_advimba.hints == {"fields": [test_advimba.name]}


async def test_can_read(test_advimba: advimba.VimbaDetector):
    # Standard detector can be used as Readable
    assert (await test_advimba.read()) == {}


async def test_decribe_describes_writer_dataset(
    test_advimba: advimba.VimbaDetector, one_shot_trigger_info: TriggerInfo
):
    assert await test_advimba.describe() == {}
    await test_advimba.stage()
    await test_advimba.prepare(one_shot_trigger_info)
    assert await test_advimba.describe() == {
        "test_advimba1": {
            "source": "mock+ca://VIMBA1:HDF1:FullFileName_RBV",
            "shape": [10, 10],
            "dtype": "array",
            "dtype_numpy": "|i1",
            "external": "STREAM:",
        }
    }


async def test_can_collect(
    test_advimba: advimba.VimbaDetector,
    static_path_provider: PathProvider,
    one_shot_trigger_info: TriggerInfo,
):
    path_info = static_path_provider()
    full_file_name = path_info.directory_path / f"{path_info.filename}.h5"

    await test_advimba.stage()
    await test_advimba.prepare(one_shot_trigger_info)
    docs = [(name, doc) async for name, doc in test_advimba.collect_asset_docs(1)]
    assert len(docs) == 2
    assert docs[0][0] == "stream_resource"
    stream_resource = docs[0][1]
    sr_uid = stream_resource["uid"]
    assert stream_resource["data_key"] == "test_advimba1"
    assert stream_resource["uri"] == "file://localhost/" + str(full_file_name).lstrip(
        "/"
    )
    assert stream_resource["parameters"] == {
        "dataset": "/entry/data/data",
        "swmr": False,
        "multiplier": 1,
        "chunk_shape": (1, 10, 10),
    }
    assert docs[1][0] == "stream_datum"
    stream_datum = docs[1][1]
    assert stream_datum["stream_resource"] == sr_uid
    assert stream_datum["seq_nums"] == {"start": 0, "stop": 0}
    assert stream_datum["indices"] == {"start": 0, "stop": 1}


async def test_can_decribe_collect(
    test_advimba: advimba.VimbaDetector, one_shot_trigger_info: TriggerInfo
):
    assert (await test_advimba.describe_collect()) == {}
    await test_advimba.stage()
    await test_advimba.prepare(one_shot_trigger_info)
    assert (await test_advimba.describe_collect()) == {
        "test_advimba1": {
            "source": "mock+ca://VIMBA1:HDF1:FullFileName_RBV",
            "shape": [10, 10],
            "dtype": "array",
            "dtype_numpy": "|i1",
            "external": "STREAM:",
        }
    }



from ._vimba import VimbaDetector
from ._vimba_controller import VimbaController
from ._vimba_io import VimbaDriverIO, VimbaExposeOutMode, VimbaOnOff, VimbaTriggerSource

__all__ = [
    "VimbaDetector",
    "VimbaController",
    "VimbaDriverIO",
    "VimbaExposeOutMode",
    "VimbaOnOff",
    "VimbaTriggerSource",
]



import pytest

from ophyd_async.sim.demo import PatternGenerator


@pytest.fixture
async def pattern_generator():
    # path: Path = tmp_path_factory.mktemp("tmp")
    pattern_generator = PatternGenerator()
    yield pattern_generator


async def test_init(pattern_generator: PatternGenerator):
    assert pattern_generator.exposure == 0.1
    assert pattern_generator.height == 240
    assert pattern_generator.width == 320
    assert pattern_generator.image_counter == 0
    assert pattern_generator._handle_for_h5_file is None
    assert pattern_generator._full_intensity_blob.shape == (240, 320)


def test_set_exposure(pattern_generator: PatternGenerator):
    pattern_generator.set_exposure(0.5)
    assert pattern_generator.exposure == 0.5


def test_set_x(pattern_generator: PatternGenerator):
    pattern_generator.set_x(5.0)
    assert pattern_generator.x == 5.0


def test_set_y(pattern_generator: PatternGenerator):
    pattern_generator.set_y(-3.0)
    assert pattern_generator.y == -3.0



import asyncio
from collections.abc import Awaitable, Callable
from unittest.mock import patch

import pytest

from ophyd_async.core import (
    DetectorTrigger,
    TriggerInfo,
    set_mock_value,
)
from ophyd_async.epics import adcore, adpilatus


@pytest.fixture
def test_adpilatus(ad_standard_det_factory) -> adpilatus.PilatusDetector:
    return ad_standard_det_factory(adpilatus.PilatusDetector)


async def test_deadtime_overridable(test_adpilatus: adpilatus.PilatusDetector):
    pilatus_controller = test_adpilatus._controller
    pilatus_controller._readout_time = adpilatus.PilatusReadoutTime.pilatus2

    # deadtime invariant with exposure time
    assert pilatus_controller.get_deadtime(0) == 2.28e-3


async def test_deadtime_invariant(
    test_adpilatus: adpilatus.PilatusDetector,
):
    pilatus_controller = test_adpilatus.controller
    # deadtime invariant with exposure time
    assert pilatus_controller.get_deadtime(0) == 0.95e-3
    assert pilatus_controller.get_deadtime(500) == 0.95e-3


@pytest.mark.parametrize(
    "detector_trigger,expected_trigger_mode",
    [
        (DetectorTrigger.internal, adpilatus.PilatusTriggerMode.internal),
        (DetectorTrigger.internal, adpilatus.PilatusTriggerMode.internal),
        (DetectorTrigger.internal, adpilatus.PilatusTriggerMode.internal),
    ],
)
async def test_trigger_mode_set(
    test_adpilatus: adpilatus.PilatusDetector,
    detector_trigger: DetectorTrigger,
    expected_trigger_mode: adpilatus.PilatusTriggerMode,
):
    async def trigger_and_complete():
        set_mock_value(test_adpilatus.drv.armed, True)
        await test_adpilatus.controller.prepare(
            TriggerInfo(number_of_triggers=1, trigger=detector_trigger)
        )
        await test_adpilatus.controller.arm()
        await test_adpilatus.controller.wait_for_idle()

    await _trigger(test_adpilatus, expected_trigger_mode, trigger_and_complete)


async def test_trigger_mode_set_without_armed_pv(
    test_adpilatus: adpilatus.PilatusDetector,
):
    async def trigger_and_complete():
        await test_adpilatus.controller.prepare(
            TriggerInfo(number_of_triggers=1, trigger=DetectorTrigger.internal)
        )
        await test_adpilatus.controller.arm()
        await test_adpilatus.controller.wait_for_idle()

    with patch(
        "ophyd_async.epics.adpilatus._pilatus_controller.DEFAULT_TIMEOUT",
        0.1,
    ):
        with pytest.raises(asyncio.TimeoutError):
            await _trigger(
                test_adpilatus,
                adpilatus.PilatusTriggerMode.internal,
                trigger_and_complete,
            )


async def _trigger(
    test_adpilatus: adpilatus.PilatusDetector,
    expected_trigger_mode: adpilatus.PilatusTriggerMode,
    trigger_and_complete: Callable[[], Awaitable],
):
    # Default TriggerMode
    assert (
        await test_adpilatus.drv.trigger_mode.get_value()
    ) == adpilatus.PilatusTriggerMode.internal

    await trigger_and_complete()

    # TriggerSource changes
    assert (await test_adpilatus.drv.trigger_mode.get_value()) == expected_trigger_mode


async def test_hints_from_hdf_writer(test_adpilatus: adpilatus.PilatusDetector):
    assert test_adpilatus.hints == {"fields": [test_adpilatus.name]}


async def test_unsupported_trigger_excepts(test_adpilatus: adpilatus.PilatusDetector):
    with pytest.raises(
        ValueError,
        # str(EnumClass.value) handling changed in Python 3.11
        match=r"PilatusController only supports the following trigger types: .* but",
    ):
        await test_adpilatus.prepare(
            TriggerInfo(
                number_of_triggers=1,
                trigger=DetectorTrigger.edge_trigger,
                deadtime=1.0,
                livetime=1.0,
            )
        )


async def test_exposure_time_and_acquire_period_set(
    test_adpilatus: adpilatus.PilatusDetector,
):
    async def dummy_open(multiplier: int = 0):
        return {}

    test_adpilatus.writer.open = dummy_open
    set_mock_value(test_adpilatus.drv.armed, True)
    await test_adpilatus.prepare(
        TriggerInfo(
            number_of_triggers=1,
            trigger=DetectorTrigger.internal,
            deadtime=1.0,
            livetime=1.0,
        )
    )
    assert (await test_adpilatus.drv.acquire_time.get_value()) == 1.0
    assert (await test_adpilatus.drv.acquire_period.get_value()) == 1.0 + 950e-6


async def test_pilatus_controller(test_adpilatus: adpilatus.PilatusDetector):
    pilatus = test_adpilatus._controller
    pilatus_driver = pilatus._drv
    set_mock_value(pilatus_driver.armed, True)
    await pilatus.prepare(
        TriggerInfo(number_of_triggers=1, trigger=DetectorTrigger.constant_gate)
    )
    await pilatus.arm()
    await pilatus.wait_for_idle()

    assert await pilatus_driver.num_images.get_value() == 1
    assert await pilatus_driver.image_mode.get_value() == adcore.ImageMode.multiple
    assert (
        await pilatus_driver.trigger_mode.get_value()
        == adpilatus.PilatusTriggerMode.ext_enable
    )
    assert await pilatus_driver.acquire.get_value() is True

    await pilatus.disarm()

    assert await pilatus_driver.acquire.get_value() is False



import re

import pytest

from ophyd_async.core import (
    DetectorTrigger,
    PathProvider,
    TriggerInfo,
    set_mock_value,
)
from ophyd_async.epics import adaravis


@pytest.fixture
def test_adaravis(ad_standard_det_factory) -> adaravis.AravisDetector:
    return ad_standard_det_factory(adaravis.AravisDetector)


@pytest.mark.parametrize("exposure_time", [0.0, 0.1, 1.0, 10.0, 100.0])
async def test_deadtime_invariant_with_exposure_time(
    exposure_time: float,
    test_adaravis: adaravis.AravisDetector,
):
    assert test_adaravis.controller.get_deadtime(exposure_time) == 1961e-6


async def test_trigger_source_set_to_gpio_line(test_adaravis: adaravis.AravisDetector):
    set_mock_value(test_adaravis.drv.trigger_source, "Freerun")

    async def trigger_and_complete():
        await test_adaravis.controller.prepare(
            TriggerInfo(
                number_of_triggers=1,
                trigger=DetectorTrigger.edge_trigger,
                livetime=None,
                deadtime=None,
                frame_timeout=None,
            )
        )
        # Prevent timeouts
        set_mock_value(test_adaravis.drv.acquire, True)

    # Default TriggerSource
    assert (await test_adaravis.drv.trigger_source.get_value()) == "Freerun"
    test_adaravis.set_external_trigger_gpio(1)
    # TriggerSource not changed by setting gpio
    assert (await test_adaravis.drv.trigger_source.get_value()) == "Freerun"

    await trigger_and_complete()

    # TriggerSource changes
    assert (await test_adaravis.drv.trigger_source.get_value()) == "Line1"

    test_adaravis.set_external_trigger_gpio(3)
    # TriggerSource not changed by setting gpio
    await trigger_and_complete()
    assert (await test_adaravis.drv.trigger_source.get_value()) == "Line3"


def test_gpio_pin_limited(test_adaravis: adaravis.AravisDetector):
    assert test_adaravis.get_external_trigger_gpio() == 1
    test_adaravis.set_external_trigger_gpio(2)
    assert test_adaravis.get_external_trigger_gpio() == 2
    with pytest.raises(
        ValueError,
        match=re.escape(
            "AravisDetector only supports the following GPIO indices: "
            "(1, 2, 3, 4) but was asked to use 55"
        ),
    ):
        test_adaravis.set_external_trigger_gpio(55)  # type: ignore


async def test_hints_from_hdf_writer(test_adaravis: adaravis.AravisDetector):
    assert test_adaravis.hints == {"fields": ["test_adaravis1"]}


async def test_can_read(test_adaravis: adaravis.AravisDetector):
    # Standard detector can be used as Readable
    assert (await test_adaravis.read()) == {}


async def test_decribe_describes_writer_dataset(
    test_adaravis: adaravis.AravisDetector, one_shot_trigger_info: TriggerInfo
):
    assert await test_adaravis.describe() == {}
    await test_adaravis.stage()
    await test_adaravis.prepare(one_shot_trigger_info)
    assert await test_adaravis.describe() == {
        "test_adaravis1": {
            "source": "mock+ca://ARAVIS1:HDF1:FullFileName_RBV",
            "shape": [10, 10],
            "dtype": "array",
            "dtype_numpy": "|i1",
            "external": "STREAM:",
        }
    }


async def test_can_collect(
    test_adaravis: adaravis.AravisDetector,
    static_path_provider: PathProvider,
    one_shot_trigger_info: TriggerInfo,
):
    path_info = static_path_provider()
    full_file_name = path_info.directory_path / f"{path_info.filename}.h5"
    await test_adaravis.stage()
    await test_adaravis.prepare(one_shot_trigger_info)
    docs = [(name, doc) async for name, doc in test_adaravis.collect_asset_docs(1)]
    assert len(docs) == 2
    assert docs[0][0] == "stream_resource"
    stream_resource = docs[0][1]
    sr_uid = stream_resource["uid"]
    assert stream_resource["data_key"] == "test_adaravis1"
    assert stream_resource["uri"] == "file://localhost/" + str(full_file_name).lstrip(
        "/"
    )
    assert stream_resource["parameters"] == {
        "dataset": "/entry/data/data",
        "swmr": False,
        "multiplier": 1,
        "chunk_shape": (1, 10, 10),
    }
    assert docs[1][0] == "stream_datum"
    stream_datum = docs[1][1]
    assert stream_datum["stream_resource"] == sr_uid
    assert stream_datum["seq_nums"] == {"start": 0, "stop": 0}
    assert stream_datum["indices"] == {"start": 0, "stop": 1}


async def test_can_decribe_collect(
    test_adaravis: adaravis.AravisDetector, one_shot_trigger_info: TriggerInfo
):
    assert (await test_adaravis.describe_collect()) == {}
    await test_adaravis.stage()
    await test_adaravis.prepare(one_shot_trigger_info)
    assert (await test_adaravis.describe_collect()) == {
        "test_adaravis1": {
            "source": "mock+ca://ARAVIS1:HDF1:FullFileName_RBV",
            "shape": [10, 10],
            "dtype": "array",
            "dtype_numpy": "|i1",
            "external": "STREAM:",
        }
    }


async def test_unsupported_trigger_excepts(test_adaravis: adaravis.AravisDetector):
    with pytest.raises(
        ValueError,
        # str(EnumClass.value) handling changed in Python 3.11
        match=r"AravisController only supports the following trigger types: .* but",
    ):
        await test_adaravis.prepare(
            TriggerInfo(
                number_of_triggers=0,
                trigger=DetectorTrigger.variable_gate,
                deadtime=1,
                livetime=1,
                frame_timeout=3,
            )
        )



from ophyd_async.core import Device, StrictEnum
from ophyd_async.epics.core import (
    epics_signal_r,
    epics_signal_rw,
    epics_signal_rw_rbv,
)

from ._utils import ADBaseDataType, FileWriteMode, ImageMode


class Callback(StrictEnum):
    Enable = "Enable"
    Disable = "Disable"


class NDArrayBaseIO(Device):
    def __init__(self, prefix: str, name: str = "") -> None:
        self.unique_id = epics_signal_r(int, prefix + "UniqueId_RBV")
        self.nd_attributes_file = epics_signal_rw(str, prefix + "NDAttributesFile")
        self.acquire = epics_signal_rw_rbv(bool, prefix + "Acquire")
        self.array_size_x = epics_signal_r(int, prefix + "ArraySizeX_RBV")
        self.array_size_y = epics_signal_r(int, prefix + "ArraySizeY_RBV")
        self.data_type = epics_signal_r(ADBaseDataType, prefix + "DataType_RBV")
        self.array_counter = epics_signal_rw_rbv(int, prefix + "ArrayCounter")
        # There is no _RBV for this one
        self.wait_for_plugins = epics_signal_rw(bool, prefix + "WaitForPlugins")
        super().__init__(name=name)


class NDPluginBaseIO(NDArrayBaseIO):
    def __init__(self, prefix: str, name: str = "") -> None:
        self.nd_array_port = epics_signal_rw_rbv(str, prefix + "NDArrayPort")
        self.enable_callbacks = epics_signal_rw_rbv(
            Callback, prefix + "EnableCallbacks"
        )
        self.nd_array_address = epics_signal_rw_rbv(int, prefix + "NDArrayAddress")
        self.array_size0 = epics_signal_r(int, prefix + "ArraySize0_RBV")
        self.array_size1 = epics_signal_r(int, prefix + "ArraySize1_RBV")
        super().__init__(prefix, name)


class NDPluginStatsIO(NDPluginBaseIO):
    """
    Plugin for computing statistics from an image or region of interest within an image.
    """

    def __init__(self, prefix: str, name: str = "") -> None:
        # Basic statistics
        self.compute_statistics = epics_signal_rw(bool, prefix + "ComputeStatistics")
        self.bgd_width = epics_signal_rw(int, prefix + "BgdWidth")
        self.total_array = epics_signal_rw(float, prefix + "TotalArray")
        # Centroid statistics
        self.compute_centroid = epics_signal_rw(bool, prefix + "ComputeCentroid")
        self.centroid_threshold = epics_signal_rw(float, prefix + "CentroidThreshold")
        # X and Y Profiles
        self.compute_profiles = epics_signal_rw(bool, prefix + "ComputeProfiles")
        self.profile_size_x = epics_signal_rw(int, prefix + "ProfileSizeX")
        self.profile_size_y = epics_signal_rw(int, prefix + "ProfileSizeY")
        self.cursor_x = epics_signal_rw(int, prefix + "CursorX")
        self.cursor_y = epics_signal_rw(int, prefix + "CursorY")
        # Array Histogram
        self.compute_histogram = epics_signal_rw(bool, prefix + "ComputeHistogram")
        self.hist_size = epics_signal_rw(int, prefix + "HistSize")
        self.hist_min = epics_signal_rw(float, prefix + "HistMin")
        self.hist_max = epics_signal_rw(float, prefix + "HistMax")
        super().__init__(prefix, name)


class DetectorState(StrictEnum):
    """
    Default set of states of an AreaDetector driver.
    See definition in ADApp/ADSrc/ADDriver.h in https://github.com/areaDetector/ADCore
    """

    Idle = "Idle"
    Acquire = "Acquire"
    Readout = "Readout"
    Correct = "Correct"
    Saving = "Saving"
    Aborting = "Aborting"
    Error = "Error"
    Waiting = "Waiting"
    Initializing = "Initializing"
    Disconnected = "Disconnected"
    Aborted = "Aborted"


class ADBaseIO(NDArrayBaseIO):
    def __init__(self, prefix: str, name: str = "") -> None:
        # Define some signals
        self.acquire_time = epics_signal_rw_rbv(float, prefix + "AcquireTime")
        self.acquire_period = epics_signal_rw_rbv(float, prefix + "AcquirePeriod")
        self.num_images = epics_signal_rw_rbv(int, prefix + "NumImages")
        self.image_mode = epics_signal_rw_rbv(ImageMode, prefix + "ImageMode")
        self.detector_state = epics_signal_r(
            DetectorState, prefix + "DetectorState_RBV"
        )
        super().__init__(prefix, name=name)


class Compression(StrictEnum):
    none = "None"
    nbit = "N-bit"
    szip = "szip"
    zlib = "zlib"
    blosc = "Blosc"
    bslz4 = "BSLZ4"
    lz4 = "LZ4"
    jpeg = "JPEG"


class NDFileHDFIO(NDPluginBaseIO):
    def __init__(self, prefix: str, name="") -> None:
        # Define some signals
        self.position_mode = epics_signal_rw_rbv(bool, prefix + "PositionMode")
        self.compression = epics_signal_rw_rbv(Compression, prefix + "Compression")
        self.num_extra_dims = epics_signal_rw_rbv(int, prefix + "NumExtraDims")
        self.file_path = epics_signal_rw_rbv(str, prefix + "FilePath")
        self.file_name = epics_signal_rw_rbv(str, prefix + "FileName")
        self.file_path_exists = epics_signal_r(bool, prefix + "FilePathExists_RBV")
        self.file_template = epics_signal_rw_rbv(str, prefix + "FileTemplate")
        self.full_file_name = epics_signal_r(str, prefix + "FullFileName_RBV")
        self.file_write_mode = epics_signal_rw_rbv(
            FileWriteMode, prefix + "FileWriteMode"
        )
        self.num_capture = epics_signal_rw_rbv(int, prefix + "NumCapture")
        self.num_captured = epics_signal_r(int, prefix + "NumCaptured_RBV")
        self.swmr_mode = epics_signal_rw_rbv(bool, prefix + "SWMRMode")
        self.lazy_open = epics_signal_rw_rbv(bool, prefix + "LazyOpen")
        self.capture = epics_signal_rw_rbv(bool, prefix + "Capture")
        self.flush_now = epics_signal_rw(bool, prefix + "FlushNow")
        self.xml_file_name = epics_signal_rw_rbv(str, prefix + "XMLFileName")
        self.array_size0 = epics_signal_r(int, prefix + "ArraySize0")
        self.array_size1 = epics_signal_r(int, prefix + "ArraySize1")
        self.create_directory = epics_signal_rw(int, prefix + "CreateDirectory")
        self.num_frames_chunks = epics_signal_r(int, prefix + "NumFramesChunks_RBV")
        self.chunk_size_auto = epics_signal_rw_rbv(bool, prefix + "ChunkSizeAuto")
        super().__init__(prefix, name)



import asyncio

from ophyd_async.core import (
    DEFAULT_TIMEOUT,
    AsyncStatus,
    DatasetDescriber,
    DetectorController,
    set_and_wait_for_value,
)
from ophyd_async.epics.adcore._utils import convert_ad_dtype_to_np

from ._core_io import ADBaseIO, DetectorState

# Default set of states that we should consider "good" i.e. the acquisition
#  is complete and went well
DEFAULT_GOOD_STATES: frozenset[DetectorState] = frozenset(
    [DetectorState.Idle, DetectorState.Aborted]
)


class ADBaseDatasetDescriber(DatasetDescriber):
    def __init__(self, driver: ADBaseIO) -> None:
        self._driver = driver

    async def np_datatype(self) -> str:
        return convert_ad_dtype_to_np(await self._driver.data_type.get_value())

    async def shape(self) -> tuple[int, int]:
        shape = await asyncio.gather(
            self._driver.array_size_y.get_value(),
            self._driver.array_size_x.get_value(),
        )
        return shape


async def set_exposure_time_and_acquire_period_if_supplied(
    controller: DetectorController,
    driver: ADBaseIO,
    exposure: float | None = None,
    timeout: float = DEFAULT_TIMEOUT,
) -> None:
    """
    Sets the exposure time if it is not None and the acquire period to the
    exposure time plus the deadtime. This is expected behavior for most
    AreaDetectors, but some may require more specialized handling.

    Parameters
    ----------
    controller:
        Controller that can supply a deadtime.
    driver:
        The driver to start acquiring. Must subclass ADBaseIO.
    exposure:
        Desired exposure time, this is a noop if it is None.
    timeout:
        How long to wait for the exposure time and acquire period to be set.
    """
    if exposure is not None:
        full_frame_time = exposure + controller.get_deadtime(exposure)
        await asyncio.gather(
            driver.acquire_time.set(exposure, timeout=timeout),
            driver.acquire_period.set(full_frame_time, timeout=timeout),
        )


async def start_acquiring_driver_and_ensure_status(
    driver: ADBaseIO,
    good_states: frozenset[DetectorState] = frozenset(DEFAULT_GOOD_STATES),
    timeout: float = DEFAULT_TIMEOUT,
) -> AsyncStatus:
    """
    Start acquiring driver, raising ValueError if the detector is in a bad state.

    This sets driver.acquire to True, and waits for it to be True up to a timeout.
    Then, it checks that the DetectorState PV is in DEFAULT_GOOD_STATES, and otherwise
    raises a ValueError.

    Parameters
    ----------
    driver:
        The driver to start acquiring. Must subclass ADBaseIO.
    good_states:
        set of states defined in DetectorState enum which are considered good states.
    timeout:
        How long to wait for driver.acquire to readback True (i.e. acquiring).

    Returns
    -------
    AsyncStatus:
        An AsyncStatus that can be awaited to set driver.acquire to True and perform
        subsequent raising (if applicable) due to detector state.
    """

    status = await set_and_wait_for_value(driver.acquire, True, timeout=timeout)

    async def complete_acquisition() -> None:
        """NOTE: possible race condition here between the callback from
        set_and_wait_for_value and the detector state updating."""
        await status
        state = await driver.detector_state.get_value()
        if state not in good_states:
            raise ValueError(
                f"Final detector state {state} not in valid end states: {good_states}"
            )

    return AsyncStatus(complete_acquisition())
